"""Syntax tree for parsed Sass expressions.

The overall structure for a Sass file uses a different kind of AST; have a look
at :mod:`scss.blockast`.
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from functools import partial
import logging
import operator

try:
    from collections import OrderedDict
except ImportError:
    # Backport
    from ordereddict import OrderedDict

import six

from scss.cssdefs import COLOR_NAMES
from scss.cssdefs import is_builtin_css_function
from scss.types import Boolean
from scss.types import Color
from scss.types import Function
from scss.types import List
from scss.types import Map
from scss.types import Null
from scss.types import String
from scss.types import Undefined
from scss.types import Url
from scss.types import Value
from scss.util import normalize_var


log = logging.getLogger(__name__)


class Expression(object):
    def __repr__(self):
        return '<%s()>' % (self.__class__.__name__)

    def evaluate(self, calculator, divide=False):
        """Evaluate this AST node, and return a Sass value.

        `divide` indicates whether a descendant node representing a division
        should be forcibly treated as a division.  See the commentary in
        `BinaryOp`.
        """
        raise NotImplementedError


class Parentheses(Expression):
    """An expression of the form `(foo)`.

    Only exists to force a slash to be interpreted as division when contained
    within parentheses.
    """
    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, repr(self.contents))

    def __init__(self, contents):
        self.contents = contents

    def evaluate(self, calculator, divide=False):
        return self.contents.evaluate(calculator, divide=True)


class UnaryOp(Expression):
    def __repr__(self):
        return '<%s(%s, %s)>' % (self.__class__.__name__, repr(self.op), repr(self.operand))

    def __init__(self, op, operand):
        self.op = op
        self.operand = operand

    def evaluate(self, calculator, divide=False):
        return self.op(self.operand.evaluate(calculator, divide=True))


class BinaryOp(Expression):
    OPERATORS = {
        operator.lt: '<',
        operator.gt: '>',
        operator.le: '<=',
        operator.ge: '>=',
        operator.eq: '==',
        operator.eq: '!=',
        operator.add: '+',
        operator.sub: '-',
        operator.mul: '*',
        operator.truediv: '/',
        operator.mod: '%',
    }

    def __repr__(self):
        return '<%s(%s, %s, %s)>' % (self.__class__.__name__, repr(self.op), repr(self.left), repr(self.right))

    def __init__(self, op, left, right):
        self.op = op
        self.left = left
        self.right = right

    def evaluate(self, calculator, divide=False):
        left = self.left.evaluate(calculator, divide=True)
        right = self.right.evaluate(calculator, divide=True)

        # Determine whether to actually evaluate, or just print the operator
        # literally.
        literal = False

        # If either operand starts with an interpolation, treat the whole
        # shebang as literal.
        if any(isinstance(operand, Interpolation) and operand.parts[0] == ''
                for operand in (self.left, self.right)):
            literal = True

        # Special handling of division: treat it as a literal slash if both
        # operands are literals, there are no parentheses, and this isn't part
        # of a bigger expression.
        # The first condition is covered by the type check.  The other two are
        # covered by the `divide` argument: other nodes that perform arithmetic
        # will pass in True, indicating that this should always be a division.
        elif (
            self.op is operator.truediv
            and not divide
            and isinstance(self.left, Literal)
            and isinstance(self.right, Literal)
        ):
            literal = True

        if literal:
            # TODO we don't currently preserve the spacing, whereas Sass
            # remembers whether there was space on either side
            op = " {0} ".format(self.OPERATORS[self.op])
            return String.unquoted(left.render() + op + right.render())

        return self.op(left, right)


class AnyOp(Expression):
    def __repr__(self):
        return '<%s(*%s)>' % (self.__class__.__name__, repr(self.operands))

    def __init__(self, *operands):
        self.operands = operands

    def evaluate(self, calculator, divide=False):
        for operand in self.operands:
            value = operand.evaluate(calculator, divide=True)
            if value:
                return value
        return value


class AllOp(Expression):
    def __repr__(self):
        return '<%s(*%s)>' % (self.__class__.__name__, repr(self.operands))

    def __init__(self, *operands):
        self.operands = operands

    def evaluate(self, calculator, divide=False):
        for operand in self.operands:
            value = operand.evaluate(calculator, divide=True)
            if not value:
                return value
        return value


class NotOp(Expression):
    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, repr(self.operand))

    def __init__(self, operand):
        self.operand = operand

    def evaluate(self, calculator, divide=False):
        operand = self.operand.evaluate(calculator, divide=True)
        return Boolean(not(operand))


class CallOp(Expression):
    def __repr__(self):
        return '<%s(%s, %s)>' % (self.__class__.__name__, repr(self.func_name), repr(self.argspec))

    def __init__(self, func_name, argspec):
        self.func_name = func_name
        self.argspec = argspec

    def evaluate(self, calculator, divide=False):
        # TODO bake this into the context and options "dicts", plus library
        func_name = normalize_var(self.func_name)

        argspec_node = self.argspec

        # Turn the pairs of arg tuples into *args and **kwargs
        # TODO unclear whether this is correct -- how does arg, kwarg, arg
        # work?
        args, kwargs = argspec_node.evaluate_call_args(calculator)
        argspec_len = len(args) + len(kwargs)

        # Translate variable names to Python identifiers
        # TODO what about duplicate kw names?  should this happen in argspec?
        # how does that affect mixins?
        kwargs = dict(
            (key.lstrip('$').replace('-', '_'), value)
            for key, value in kwargs.items())

        # TODO merge this with the library
        funct = None
        try:
            funct = calculator.namespace.function(func_name, argspec_len)
        except KeyError:
            try:
                # DEVIATION: Fall back to single parameter
                funct = calculator.namespace.function(func_name, 1)
                args = [List(args, use_comma=True)]
            except KeyError:
                if not is_builtin_css_function(func_name):
                    log.error("Function not found: %s:%s", func_name, argspec_len, extra={'stack': True})

        if funct:
            if getattr(funct, '_pyscss_needs_namespace', False):
                # @functions and some Python functions take the namespace as an
                # extra first argument
                ret = funct(calculator.namespace, *args, **kwargs)
            else:
                ret = funct(*args, **kwargs)
            if not isinstance(ret, Value):
                raise TypeError("Expected Sass type as return value, got %r" % (ret,))
            return ret

        # No matching function found, so render the computed values as a CSS
        # function call.  Slurpy arguments are expanded and named arguments are
        # unsupported.
        if kwargs:
            raise TypeError("The CSS function %s doesn't support keyword arguments." % (func_name,))

        # TODO another candidate for a "function call" sass type
        rendered_args = [arg.render() for arg in args]

        return String(
            "%s(%s)" % (func_name, ", ".join(rendered_args)),
            quotes=None)


# TODO this class should delegate the unescaping to the type, rather than
# burying it in the parser
class Interpolation(Expression):
    """A string that may contain any number of interpolations:

        foo#{...}bar#{...}baz
    """
    def __init__(self, parts, quotes=None, type=String, **kwargs):
        self.parts = parts
        self.quotes = quotes
        self.type = type
        self.kwargs = kwargs

    def __repr__(self):
        repr_parts = []
        for i, part in enumerate(self.parts):
            if i % 2 == 0:
                if part:
                    repr_parts.append(repr(part))
            else:
                repr_parts.append('#{' + repr(part) + '}')

        return "<{0} {1}>".format(type(self).__name__, " ".join(repr_parts))

    @classmethod
    def maybe(cls, parts, quotes=None, type=String, **kwargs):
        """Returns an interpolation if there are multiple parts, otherwise a
        plain Literal.  This keeps the AST somewhat simpler, but also is the
        only way `Literal.from_bareword` gets called.
        """
        if len(parts) > 1:
            return cls(parts, quotes=quotes, type=type, **kwargs)

        if quotes is None and type is String:
            return Literal.from_bareword(parts[0])

        return Literal(type(parts[0], quotes=quotes, **kwargs))

    def evaluate(self, calculator, divide=False):
        result = []
        for i, part in enumerate(self.parts):
            if i % 2 == 0:
                # First part and other odd parts are literal string
                result.append(part)
            else:
                # Interspersed (even) parts are nodes
                value = part.evaluate(calculator, divide)
                # TODO need to know whether to pass `compress` here
                result.append(value.render_interpolated())

        return self.type(''.join(result), quotes=self.quotes, **self.kwargs)



class Literal(Expression):
    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, repr(self.value))

    def __init__(self, value):
        self.value = value

    @classmethod
    def from_bareword(cls, word):
        if word in COLOR_NAMES:
            value = Color.from_name(word)
        elif word == 'null':
            value = Null()
        elif word == 'undefined':
            value = Undefined()
        elif word == 'true':
            value = Boolean(True)
        elif word == 'false':
            value = Boolean(False)
        else:
            value = String(word, quotes=None)

        return cls(value)

    def evaluate(self, calculator, divide=False):
        if (isinstance(self.value, Undefined) and
                calculator.undefined_variables_fatal):
            raise SyntaxError("Undefined literal.")

        return self.value


class Variable(Expression):
    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, repr(self.name))

    def __init__(self, name):
        self.name = name

    def evaluate(self, calculator, divide=False):
        try:
            value = calculator.namespace.variable(self.name)
        except KeyError:
            if calculator.undefined_variables_fatal:
                raise SyntaxError("Undefined variable: '%s'." % self.name)
            else:
                log.error("Undefined variable '%s'", self.name, extra={'stack': True})
                return Undefined()
        else:
            if isinstance(value, six.string_types):
                log.warn(
                    "Expected a Sass type for the value of {0}, "
                    "but found a string expression: {1!r}"
                    .format(self.name, value)
                )
                evald = calculator.evaluate_expression(value)
                if evald is not None:
                    return evald
            return value


class ListLiteral(Expression):
    def __repr__(self):
        return '<%s(%s, comma=%s)>' % (self.__class__.__name__, repr(self.items), repr(self.comma))

    def __init__(self, items, comma=True):
        self.items = items
        self.comma = comma

    def evaluate(self, calculator, divide=False):
        items = [item.evaluate(calculator, divide=divide) for item in self.items]

        # Whether this is a "plain" literal matters for null removal: nulls are
        # left alone if this is a completely vanilla CSS property
        literal = True
        if divide:
            # TODO sort of overloading "divide" here...  rename i think
            literal = False
        elif not all(isinstance(item, Literal) for item in self.items):
            literal = False

        return List(items, use_comma=self.comma, literal=literal)


class MapLiteral(Expression):
    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, repr(self.pairs))

    def __init__(self, pairs):
        self.pairs = tuple((var, value) for var, value in pairs if value is not None)

    def evaluate(self, calculator, divide=False):
        scss_pairs = []
        for key, value in self.pairs:
            scss_pairs.append((
                key.evaluate(calculator),
                value.evaluate(calculator),
            ))

        return Map(scss_pairs)


class ArgspecLiteral(Expression):
    """Contains pairs of argument names and values, as parsed from a function
    definition or function call.

    Note that the semantics are somewhat ambiguous.  Consider parsing:

        $foo, $bar: 3

    If this appeared in a function call, $foo would refer to a value; if it
    appeared in a function definition, $foo would refer to an existing
    variable.  This it's up to the caller to use the right iteration function.
    """
    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, repr(self.argpairs))

    def __init__(self, argpairs, slurp=None):
        # argpairs is a list of 2-tuples, parsed as though this were a function
        # call, so (variable name as string or None, default value as AST
        # node).
        # slurp is the name of a variable to receive slurpy arguments.
        self.argpairs = tuple(argpairs)
        if slurp is all:
            # DEVIATION: special syntax to allow injecting arbitrary arguments
            # from the caller to the callee
            self.inject = True
            self.slurp = None
        elif slurp:
            self.inject = False
            self.slurp = Variable(slurp)
        else:
            self.inject = False
            self.slurp = None

    def iter_list_argspec(self):
        yield None, ListLiteral(zip(*self.argpairs)[1])

    def iter_def_argspec(self):
        """Interpreting this literal as a function definition, yields pairs of
        (variable name as a string, default value as an AST node or None).
        """
        started_kwargs = False
        seen_vars = set()

        for var, value in self.argpairs:
            if var is None:
                # value is actually the name
                var = value
                value = None

                if started_kwargs:
                    raise SyntaxError(
                        "Required argument %r must precede optional arguments"
                        % (var.name,))

            else:
                started_kwargs = True

            if not isinstance(var, Variable):
                raise SyntaxError("Expected variable name, got %r" % (var,))

            if var.name in seen_vars:
                raise SyntaxError("Duplicate argument %r" % (var.name,))
            seen_vars.add(var.name)

            yield var.name, value

    def evaluate_call_args(self, calculator):
        """Interpreting this literal as a function call, return a 2-tuple of
        ``(args, kwargs)``.
        """
        args = []
        kwargs = OrderedDict()  # Sass kwargs preserve order
        for var_node, value_node in self.argpairs:
            value = value_node.evaluate(calculator, divide=True)
            if var_node is None:
                # Positional
                args.append(value)
            else:
                # Named
                if not isinstance(var_node, Variable):
                    raise TypeError(
                        "Expected variable name, got {0!r}".format(var_node))
                kwargs[var_node.name] = value

        # Slurpy arguments go on the end of the args
        if self.slurp:
            args.extend(self.slurp.evaluate(calculator, divide=True))

        return args, kwargs


class FunctionLiteral(Expression):
    """Wraps an existing AST node in a literal (unevaluated) function call."""
    def __init__(self, child, function_name):
        self.child = child
        self.function_name = function_name

    def evaluate(self, calculator, divide=False):
        child = self.child.evaluate(calculator, divide)
        if isinstance(child, String):
            contents = child.value
            quotes = child.quotes
        else:
            # TODO compress
            contents = child.render()
            quotes = None

        # TODO unclear if this is the right place for this logic, or if it
        # should go in the Function constructor, or should be passed in
        # explicitly by the grammar, or even if Url should go away entirely
        if self.function_name == "url":
            return Url(contents, quotes=quotes)
        else:
            return Function(contents, self.function_name, quotes=quotes)


class AlphaFunctionLiteral(Expression):
    """Wraps an existing AST node in a literal (unevaluated) function call,
    prepending "opacity=" to the contents.
    """
    def __init__(self, child):
        self.child = child

    def evaluate(self, calculator, divide=False):
        child = self.child.evaluate(calculator, divide)
        if isinstance(child, String):
            contents = child.value
        else:
            # TODO compress
            contents = child.render()
        return Function('opacity=' + contents, 'alpha', quotes=None)


class TernaryOp(Expression):
    """Sass implements this with a function:

        prop: if(condition, true-value, false-value);

    However, the second and third arguments are guaranteed not to be evaluated
    unless necessary.  Functions always receive evaluated arguments, so this is
    a syntactic construct in disguise.
    """
    def __repr__(self):
        return '<%s(%r, %r, %r)>' % (
            self.__class__.__name__,
            self.condition,
            self.true_expression,
            self.false_expression,
        )

    def __init__(self, list_literal):
        args = list_literal.items
        if len(args) != 3:
            raise SyntaxError("if() must have exactly 3 arguments")
        self.condition, self.true_expression, self.false_expression = args

    def evaluate(self, calculator, divide=False):
        if self.condition.evaluate(calculator, divide=True):
            return self.true_expression.evaluate(calculator, divide=True)
        else:
            return self.false_expression.evaluate(calculator, divide=True)



# -*- coding: utf-8 -*-
"""
Tool for converting Less files to Scss

Usage: python -m scss.less2scss [file]

"""
# http://stackoverflow.com/questions/14970224/anyone-know-of-a-good-way-to-convert-from-less-to-sass
from __future__ import unicode_literals, absolute_import, print_function

import re
import os
import sys


class Less2Scss(object):
    at_re = re.compile(r'@(?!(media|import|mixin|font-face)(\s|\())')
    mixin_re = re.compile(r'\.([\w\-]*)\s*\((.*)\)\s*\{')
    include_re = re.compile(r'(\s|^)\.([\w\-]*\(?.*\)?;)')
    functions_map = {
        'spin': 'adjust-hue',
    }
    functions_re = re.compile(r'(%s)\(' % '|'.join(functions_map))

    def convert(self, content):
        content = self.convertVariables(content)
        content = self.convertMixins(content)
        content = self.includeMixins(content)
        content = self.convertFunctions(content)
        return content

    def convertVariables(self, content):
        # Matches any @ that doesn't have 'media ' or 'import ' after it.
        content = self.at_re.sub('$', content)
        return content

    def convertMixins(self, content):
        content = self.mixin_re.sub('@mixin \1(\2) {', content)
        return content

    def includeMixins(self, content):
        content = self.mixin_re.sub('\1@include \2', content)
        return content

    def convertFunctions(self, content):
        content = self.functions_re.sub(lambda m: '%s(' % self.functions_map[m.group(0)], content)
        return content


def less2scss(options, args):
    if not args:
        args = ['-']

    less2scss = Less2Scss()

    for source_path in args:
        if source_path == '-':
            source = sys.stdin
            destiny = sys.stdout
        else:
            try:
                source = open(source_path)
                destiny_path, ext = os.path.splitext(source_path)
                destiny_path += '.scss'
                if not options.force and os.path.exists(destiny_path):
                    raise IOError("File already exists: %s" % destiny_path)
                destiny = open(destiny_path, 'w')
            except Exception as e:
                error = "%s" % e
                if destiny_path in error:
                    ignoring = "Ignoring"
                else:
                    ignoring = "Ignoring %s" % destiny_path
                print("WARNING -- %s. %s" % (ignoring, error), file=sys.stderr)
                continue
        content = source.read()
        content = less2scss.convert(content)
        destiny.write(content)


def main():
    from optparse import OptionParser, SUPPRESS_HELP

    parser = OptionParser(usage="Usage: %prog [file]",
                          description="Converts Less files to Scss.",
                          add_help_option=False)
    parser.add_option("-f", "--force", action="store_true",
                      dest="force", default=False,
                      help="Forces overwriting output file if it already exists")
    parser.add_option("-?", action="help", help=SUPPRESS_HELP)
    parser.add_option("-h", "--help", action="help",
                      help="Show this message and exit")
    parser.add_option("-v", "--version", action="store_true",
                      help="Print version and exit")

    options, args = parser.parse_args()

    if options.version:
        from scss.tool import print_version
        print_version()
    else:
        less2scss(options, args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Run me with the path to a sassc checkout, and I'll copy over all the tests
that are viable to copy.
"""

# TODO:
# - huge.scss takes ages to run
# - extraneous files:
#   - spec/scss/guard_assign  (sassc_output.css, sass_output.css)
#   - spec/basic/14_imports  (a.scss, b.scss, d.scss, sub)
#   - spec/basic/12_pseudo_classes_and_elements  (notes.txt)
#   - spec/todo/options_passed_to_script  (options.cfg)
#   - spec/todo/newlines_removed_from_selectors_when_compressed  (options.cfg)
# - files that ruby sass fails to compile:
#   - spec/todo/variables_in_media.scss
#   - spec/todo/css_error_with_windows_newlines.scss
#   - tests/charset.scss
#   - tests/warnings.scss
#   - tests/another-gradient-test.scss
#   - tests/media.scss
#   - tests/empty-properties.scss
#   - tests/for_in_function.scss
#   - tests/directives-in-propsets.scss
#   - tests/trace.scss
#   - tests/empty-rule.scss


import os
import os.path
import shutil
import subprocess
import sys

def compile_with_ruby_sass(infile):
    outfile = os.path.splitext(infile)[0] + '.css'

    with open(outfile, 'w') as outf:
        # "expanded" is pretty close to what pyscss spits out, but it
        # includes extra blank lines between nested blocks, so remove those
        # with a quick grep
        scss_proc = subprocess.Popen(['scss', '--style=expanded', infile], stdout=subprocess.PIPE)
        grep_proc = subprocess.Popen(['grep', '.'], stdin=scss_proc.stdout, stdout=outf)
        ret = grep_proc.wait()

    if ret != 0:
        print("...whoops, bailing on {}".format(infile))
        os.unlink(infile)
        try:
            os.unlink(outfile)
        except OSError:
            pass

def sync_tests(sassc_root):
    target_dir = os.path.dirname(__file__)

    # Grab all the input/output pairs from `spec`
    spec_dir = os.path.join(sassc_root, 'spec')
    for dirpath, dirnames, filenames in os.walk(spec_dir):
        fileset = set(filenames)
        try:
            fileset.discard('input.scss')
            infile = 'input.scss'
        except KeyError:
            continue

        path = os.path.relpath(dirpath, sassc_root)
        if 'output.css' in fileset:
            outfile = 'output.css'
        elif 'expected_output.css' in fileset:
            outfile = 'expected_output.css'
        else:
            print("can't find an output file in {}, skipping".format(path))
            continue
        fileset.discard(outfile)

        test_name = path.replace('/', '-')
        print("cool, found a test pair:", path)

        if fileset or dirnames:
            # TODO
            print("... extra files i don't know how to handle, skipping -- {!r} {!r}".format(fileset, dirnames))
            continue

        shutil.copyfile(
            os.path.join(dirpath, infile),
            os.path.join(target_dir, test_name + '.scss'))

        compile_with_ruby_sass(os.path.join(target_dir, test_name + '.scss'))

    # Grab the extra tests from `tests` (but don't recurse)
    tests_dir = os.path.join(sassc_root, 'tests')
    for fn in os.listdir(tests_dir):
        path = os.path.join(tests_dir, fn)
        if not os.path.isfile(path) or not path.endswith('.scss'):
            continue

        target = os.path.join(target_dir, 'tests-' + fn)
        shutil.copyfile(path, target)
        compile_with_ruby_sass(target)




if __name__ == '__main__':
    sync_tests(*sys.argv[1:])



from __future__ import absolute_import
from __future__ import unicode_literals

from scss import Scss

import pytest

# py.test bug: unicode literals not allowed here, so cast to native str type
pytestmark = pytest.mark.skipif(str("not config.getoption('include_ruby')"))

# TODO undupe
def assert_rendering(input, expected, **kwargs):
    compiler = Scss(scss_opts=dict(compress=False), **kwargs)
    css = compiler.compile(input)

    # TODO chumptastic hack; sass and pyscss have slightly different
    # "non-compressed" output
    import re
    css = re.sub(r'(?m)\n *[}]$', ' }\n', css).rstrip("\n") + "\n"
    #css = re.sub(r'; [}]', ';\n  }', css)
    #css = re.sub(r'\n *[}]$', ' }', css)

    assert expected == css

def test_basic():
    assert_rendering('''\
.foo {a: b}
.bar {@extend .foo}
''', '''\
.foo, .bar {
  a: b; }
''')

    assert_rendering('''\
.bar {@extend .foo}
.foo {a: b}
''', '''\
.foo, .bar {
  a: b; }
''')

    assert_rendering('''\
.foo {a: b}
.bar {c: d; @extend .foo}
''', '''\
.foo, .bar {
  a: b; }

.bar {
  c: d; }
''')

    assert_rendering('''\
.foo {a: b}
.bar {@extend .foo; c: d}
''', '''\
.foo, .bar {
  a: b; }

.bar {
  c: d; }
''')

def test_multiple_targets():
    assert_rendering('''\
.foo {a: b}
.bar {@extend .foo}
.blip .foo {c: d}
''', '''\
.foo, .bar {
  a: b; }

.blip .foo, .blip .bar {
  c: d; }
''')

def test_multiple_extendees():
    assert_rendering('''\
.foo {a: b}
.bar {c: d}
.baz {@extend .foo; @extend .bar}
''', '''\
.foo, .baz {
  a: b; }

.bar, .baz {
  c: d; }
''')

def test_multiple_extends_with_single_extender_and_single_target():
    assert_extends(
        '.foo .bar',
        '.baz {@extend .foo; @extend .bar}',
        '.foo .bar, .baz .bar, .foo .baz, .baz .baz')
    assert_extends(
        '.foo.bar',
        '.baz {@extend .foo; @extend .bar}',
        '.foo.bar, .baz')

def test_multiple_extends_with_multiple_extenders_and_single_target():
    assert_rendering('''\
.foo .bar {a: b}
.baz {@extend .foo}
.bang {@extend .bar}
''', '''\
.foo .bar, .baz .bar, .foo .bang, .baz .bang {
  a: b; }
''')

    assert_rendering('''\
.foo.bar {a: b}
.baz {@extend .foo}
.bang {@extend .bar}
''', '''\
.foo.bar, .bar.baz, .baz.bang, .foo.bang {
  a: b; }
''')

def test_chained_extends():
    assert_rendering('''\
.foo {a: b}
.bar {@extend .foo}
.baz {@extend .bar}
.bip {@extend .bar}
''', '''\
.foo, .bar, .baz, .bip {
  a: b; }
''')

def test_dynamic_extendee():
    assert_extends(
        '.foo',
        '.bar {@extend #{".foo"}}',
        '.foo, .bar')
    assert_extends(
        '[baz^="blip12px"]',
        '.bar {@extend [baz^="blip#{12px}"]}',
        '[baz^="blip12px"], .bar')

def test_nested_target():
    assert_extends(
        '.foo .bar',
        '.baz {@extend .bar}',
        '.foo .bar, .foo .baz')

def test_target_with_child():
    assert_extends(
        '.foo .bar',
        '.baz {@extend .foo}',
        '.foo .bar, .baz .bar')

def test_class_unification():
    assert_unification(
        '.foo.bar',
        '.baz {@extend .foo}',
        '.foo.bar, .bar.baz')
    assert_unification(
        '.foo.baz',
        '.baz {@extend .foo}',
        '.baz')

def test_id_unification():
    assert_unification(
        '.foo.bar',
        '#baz {@extend .foo}',
        '.foo.bar, .bar#baz')
    assert_unification(
        '.foo#baz',
        '#baz {@extend .foo}',
        '#baz')

    # XXX assert_extend_doesnt_match('#bar', '.foo', :failed_to_unify, 2) do
    assert_unification(
        '.foo#baz',
        '#bar {@extend .foo}',
        '.foo#baz')

def test_universal_unification_with_simple_target():
    assert_unification(
        '.foo',
        '* {@extend .foo}',
        '.foo, *')
    assert_unification(
        '.foo',
        '*|* {@extend .foo}',
        '.foo, *|*')
    assert_unification(
        '.foo.bar',
        '* {@extend .foo}',
        '.bar')
    assert_unification(
        '.foo.bar',
        '*|* {@extend .foo}',
        '.bar')
    assert_unification(
        '.foo.bar',
        'ns|* {@extend .foo}',
        '.foo.bar, ns|*.bar')

def test_universal_unification_with_namespaceless_universal_target():
    assert_unification(
        '*.foo',
        '* {@extend .foo}',
        '*')
    assert_unification(
        '*.foo',
        '*|* {@extend .foo}',
        '*')
    assert_unification(
        '*|*.foo',
        '* {@extend .foo}',
        '*|*.foo, *')
    assert_unification(
        '*|*.foo',
        '*|* {@extend .foo}',
        '*|*')
    assert_unification(
        '*.foo',
        'ns|* {@extend .foo}',
        '*.foo, ns|*')
    assert_unification(
        '*|*.foo',
        'ns|* {@extend .foo}',
        '*|*.foo, ns|*')

def test_universal_unification_with_namespaced_universal_target():
    assert_unification(
        'ns|*.foo',
        '* {@extend .foo}',
        'ns|*')
    assert_unification(
        'ns|*.foo',
        '*|* {@extend .foo}',
        'ns|*')

    # XXX assert_extend_doesnt_match('ns2|*', '.foo', :failed_to_unify, 2) do
    assert_unification(
        'ns1|*.foo',
        'ns2|* {@extend .foo}',
        'ns1|*.foo')

    assert_unification(
        'ns|*.foo',
        'ns|* {@extend .foo}',
        'ns|*')

def test_universal_unification_with_namespaceless_element_target():
    assert_unification(
        'a.foo',
        '* {@extend .foo}',
        'a')
    assert_unification(
        'a.foo',
        '*|* {@extend .foo}',
        'a')
    assert_unification(
        '*|a.foo',
        '* {@extend .foo}',
        '*|a.foo, a')
    assert_unification(
        '*|a.foo',
        '*|* {@extend .foo}',
        '*|a')
    assert_unification(
        'a.foo',
        'ns|* {@extend .foo}',
        'a.foo, ns|a')
    assert_unification(
        '*|a.foo',
        'ns|* {@extend .foo}',
        '*|a.foo, ns|a')

def test_universal_unification_with_namespaced_element_target():
    assert_unification(
        'ns|a.foo',
        '* {@extend .foo}',
        'ns|a')
    assert_unification(
        'ns|a.foo',
        '*|* {@extend .foo}',
        'ns|a')

    # XXX assert_extend_doesnt_match('ns2|*', '.foo', :failed_to_unify, 2) do
    assert_unification(
        'ns1|a.foo',
        'ns2|* {@extend .foo}',
        'ns1|a.foo')

    assert_unification(
        'ns|a.foo',
        'ns|* {@extend .foo}',
        'ns|a')

def test_element_unification_with_simple_target():
    assert_unification(
        '.foo',
        'a {@extend .foo}',
        '.foo, a')
    assert_unification(
        '.foo.bar',
        'a {@extend .foo}',
        '.foo.bar, a.bar')
    assert_unification(
        '.foo.bar',
        '*|a {@extend .foo}',
        '.foo.bar, *|a.bar')
    assert_unification(
        '.foo.bar',
        'ns|a {@extend .foo}',
        '.foo.bar, ns|a.bar')

def test_element_unification_with_namespaceless_universal_target():
    assert_unification(
        '*.foo',
        'a {@extend .foo}',
        '*.foo, a')
    assert_unification(
        '*.foo',
        '*|a {@extend .foo}',
        '*.foo, a')
    assert_unification(
        '*|*.foo',
        'a {@extend .foo}',
        '*|*.foo, a')
    assert_unification(
        '*|*.foo',
        '*|a {@extend .foo}',
        '*|*.foo, *|a')
    assert_unification(
        '*.foo',
        'ns|a {@extend .foo}',
        '*.foo, ns|a')
    assert_unification(
        '*|*.foo',
        'ns|a {@extend .foo}',
        '*|*.foo, ns|a')

def test_element_unification_with_namespaced_universal_target():
    assert_unification(
        'ns|*.foo',
        'a {@extend .foo}',
        'ns|*.foo, ns|a')
    assert_unification(
        'ns|*.foo',
        '*|a {@extend .foo}',
        'ns|*.foo, ns|a')

    # XXX assert_extend_doesnt_match('ns2|a', '.foo', :failed_to_unify, 2) do
    assert_unification(
        'ns1|*.foo',
        'ns2|a {@extend .foo}',
        'ns1|*.foo')

    assert_unification(
        'ns|*.foo',
        'ns|a {@extend .foo}',
        'ns|*.foo, ns|a')

def test_element_unification_with_namespaceless_element_target():
    assert_unification(
        'a.foo',
        'a {@extend .foo}',
        'a')
    assert_unification(
        'a.foo',
        '*|a {@extend .foo}',
        'a')
    assert_unification(
        '*|a.foo',
        'a {@extend .foo}',
        '*|a.foo, a')
    assert_unification(
        '*|a.foo',
        '*|a {@extend .foo}',
        '*|a')
    assert_unification(
        'a.foo',
        'ns|a {@extend .foo}',
        'a.foo, ns|a')
    assert_unification(
        '*|a.foo',
        'ns|a {@extend .foo}',
        '*|a.foo, ns|a')

    # XXX assert_extend_doesnt_match('h1', '.foo', :failed_to_unify, 2) do
    assert_unification(
        'a.foo',
        'h1 {@extend .foo}',
        'a.foo')

def test_element_unification_with_namespaced_element_target():
    assert_unification(
        'ns|a.foo',
        'a {@extend .foo}',
        'ns|a')
    assert_unification(
        'ns|a.foo',
        '*|a {@extend .foo}',
        'ns|a')

    # XXX assert_extend_doesnt_match('ns2|a', '.foo', :failed_to_unify, 2) do
    assert_unification(
        'ns1|a.foo',
        'ns2|a {@extend .foo}',
        'ns1|a.foo')

    assert_unification(
        'ns|a.foo',
        'ns|a {@extend .foo}',
        'ns|a')

def test_attribute_unification():
    assert_unification(
        '[foo=bar].baz',
        '[foo=baz] {@extend .baz}',
        '[foo=bar].baz, [foo=bar][foo=baz]')
    assert_unification(
        '[foo=bar].baz',
        '[foo^=bar] {@extend .baz}',
        '[foo=bar].baz, [foo=bar][foo^=bar]')
    assert_unification(
        '[foo=bar].baz',
        '[foot=bar] {@extend .baz}',
        '[foo=bar].baz, [foo=bar][foot=bar]')
    assert_unification(
        '[foo=bar].baz',
        '[ns|foo=bar] {@extend .baz}',
        '[foo=bar].baz, [foo=bar][ns|foo=bar]')
    assert_unification(
        '%-a [foo=bar].bar',
        '[foo=bar] {@extend .bar}',
        '-a [foo=bar]')

def test_pseudo_unification():
    assert_unification(
        ':foo.baz',
        ':foo(2n+1) {@extend .baz}',
        ':foo.baz, :foo:foo(2n+1)')
    assert_unification(
        ':foo.baz',
        '::foo {@extend .baz}',
        ':foo.baz, :foo::foo')

    # XXX assert_extend_doesnt_match('::bar', '.baz', :failed_to_unify, 2) do
    assert_unification(
        '::foo.baz',
        '::bar {@extend .baz}',
        '::foo.baz')

    # XXX assert_extend_doesnt_match('::foo(2n+1)', '.baz', :failed_to_unify, 2) do
    assert_unification(
        '::foo.baz',
        '::foo(2n+1) {@extend .baz}',
        '::foo.baz')

    assert_unification(
        '::foo.baz',
        '::foo {@extend .baz}',
        '::foo')
    assert_unification(
        '::foo(2n+1).baz',
        '::foo(2n+1) {@extend .baz}',
        '::foo(2n+1)')
    assert_unification(
        ':foo.baz',
        ':bar {@extend .baz}',
        ':foo.baz, :foo:bar')
    assert_unification(
        '.baz:foo',
        ':after {@extend .baz}',
        '.baz:foo, :foo:after')
    assert_unification(
        '.baz:after',
        ':foo {@extend .baz}',
        '.baz:after, :foo:after')
    assert_unification(
        ':foo.baz',
        ':foo {@extend .baz}',
        ':foo')

def test_pseudoelement_remains_at_end_of_selector():
    assert_extends(
        '.foo::bar',
        '.baz {@extend .foo}',
        '.foo::bar, .baz::bar')
    assert_extends(
        'a.foo::bar',
        '.baz {@extend .foo}',
        'a.foo::bar, a.baz::bar')

def test_pseudoclass_remains_at_end_of_selector():
    assert_extends(
        '.foo:bar',
        '.baz {@extend .foo}',
        '.foo:bar, .baz:bar')
    assert_extends(
        'a.foo:bar',
        '.baz {@extend .foo}',
        'a.foo:bar, a.baz:bar')

def test_not_remains_at_end_of_selector():
    assert_extends(
        '.foo:not(.bar)',
        '.baz {@extend .foo}',
        '.foo:not(.bar), .baz:not(.bar)')

def test_pseudoelement_goes_lefter_than_pseudoclass():
    assert_extends(
        '.foo::bar',
        '.baz:bang {@extend .foo}',
        '.foo::bar, .baz:bang::bar')
    assert_extends(
        '.foo:bar',
        '.baz::bang {@extend .foo}',
        '.foo:bar, .baz:bar::bang')

def test_pseudoelement_goes_lefter_than_not():
    assert_extends(
        '.foo::bar',
        '.baz:not(.bang) {@extend .foo}',
        '.foo::bar, .baz:not(.bang)::bar')
    assert_extends(
        '.foo:not(.bang)',
        '.baz::bar {@extend .foo}',
        '.foo:not(.bang), .baz:not(.bang)::bar')

def test_negation_unification():
    assert_unification(
        ':not(.foo).baz',
        ':not(.bar) {@extend .baz}',
        ':not(.foo).baz, :not(.foo):not(.bar)')
    assert_unification(
        ':not(.foo).baz',
        ':not(.foo) {@extend .baz}',
        ':not(.foo)')
    assert_unification(
        ':not([a=b]).baz',
        ':not([a = b]) {@extend .baz}',
        ':not([a=b])')

def test_comma_extendee():
    assert_rendering('''\
.foo {a: b}
.bar {c: d}
.baz {@extend .foo, .bar}
''', '''\
.foo, .baz {
  a: b; }

.bar, .baz {
  c: d; }
''')

def test_redundant_selector_elimination():
    assert_rendering('''\
.foo.bar {a: b}
.x {@extend .foo, .bar}
.y {@extend .foo, .bar}
''', '''\
.foo.bar, .x, .y {
  a: b; }
''')

  ## Long Extendees

def test_long_extendee():
    assert_extends(
        '.foo.bar',
        '.baz {@extend .foo.bar}',
        '.foo.bar, .baz')

def test_long_extendee_requires_all_selectors():
    # XXX assert_extend_doesnt_match('.baz', '.foo.bar', :not_found, 2) do
    assert_extends(
        '.foo',
        '.baz {@extend .foo.bar}',
        '.foo')

def test_long_extendee_matches_supersets():
    assert_extends(
        '.foo.bar.bap',
        '.baz {@extend .foo.bar}',
        '.foo.bar.bap, .bap.baz')

def test_long_extendee_runs_unification():
    assert_extends(
        'ns|*.foo.bar',
        'a.baz {@extend .foo.bar}',
        'ns|*.foo.bar, ns|a.baz')

  ## Long Extenders

def test_long_extender():
    assert_extends(
        '.foo.bar',
        '.baz.bang {@extend .foo}',
        '.foo.bar, .bar.baz.bang')

def test_long_extender_runs_unification():
    assert_extends(
        'ns|*.foo.bar',
        'a.baz {@extend .foo}',
        'ns|*.foo.bar, ns|a.bar.baz')

def test_long_extender_aborts_unification():
    # XXX assert_extend_doesnt_match('h1.baz', '.foo', :failed_to_unify, 2) do
    assert_extends(
        'a.foo#bar',
        'h1.baz {@extend .foo}',
        'a.foo#bar')

    # XXX assert_extend_doesnt_match('.bang#baz', '.foo', :failed_to_unify, 2) do
    assert_extends(
        'a.foo#bar',
        '.bang#baz {@extend .foo}',
        'a.foo#bar')

  ## Nested Extenders

def test_nested_extender():
    assert_extends(
        '.foo',
        'foo bar {@extend .foo}',
        '.foo, foo bar')

def test_nested_extender_runs_unification():
    assert_extends(
        '.foo.bar',
        'foo bar {@extend .foo}',
        '.foo.bar, foo bar.bar')

def test_nested_extender_aborts_unification():
    # XXX assert_extend_doesnt_match('foo bar', '.foo', :failed_to_unify, 2) do
    assert_extends(
        'baz.foo',
        'foo bar {@extend .foo}',
        'baz.foo')

def test_nested_extender_alternates_parents():
    assert_extends('.baz .bip .foo', 'foo .grank bar {@extend .foo}',
      '.baz .bip .foo, .baz .bip foo .grank bar, foo .grank .baz .bip bar')

def test_nested_extender_unifies_identical_parents():
    assert_extends('.baz .bip .foo', '.baz .bip bar {@extend .foo}',
      '.baz .bip .foo, .baz .bip bar')

def test_nested_extender_unifies_common_substring():
    assert_extends('.baz .bip .bap .bink .foo', '.brat .bip .bap bar {@extend .foo}',
      '.baz .bip .bap .bink .foo, .baz .brat .bip .bap .bink bar, .brat .baz .bip .bap .bink bar')

def test_nested_extender_unifies_common_subseq():
    assert_extends('.a .x .b .y .foo', '.a .n .b .m bar {@extend .foo}',
      '.a .x .b .y .foo, .a .x .n .b .y .m bar, .a .n .x .b .y .m bar, .a .x .n .b .m .y bar, .a .n .x .b .m .y bar')

def test_nested_extender_chooses_first_subseq():
    assert_extends('.a .b .c .d .foo', '.c .d .a .b .bar {@extend .foo}',
      '.a .b .c .d .foo, .a .b .c .d .a .b .bar')

def test_nested_extender_counts_extended_subselectors():
    assert_extends('.a .bip.bop .foo', '.b .bip .bar {@extend .foo}',
      '.a .bip.bop .foo, .a .b .bip.bop .bar, .b .a .bip.bop .bar')

def test_nested_extender_counts_extended_superselectors():
    assert_extends('.a .bip .foo', '.b .bip.bop .bar {@extend .foo}',
      '.a .bip .foo, .a .b .bip.bop .bar, .b .a .bip.bop .bar')

def test_nested_extender_with_child_selector():
    assert_extends(
        '.baz .foo',
        'foo > bar {@extend .foo}',
        '.baz .foo, .baz foo > bar')

def test_nested_extender_finds_common_selectors_around_child_selector():
    assert_extends(
        'a > b c .c1',
        'a c .c2 {@extend .c1}',
        'a > b c .c1, a > b c .c2')
    assert_extends(
        'a > b c .c1',
        'b c .c2 {@extend .c1}',
        'a > b c .c1, a > b c .c2')

def test_nested_extender_doesnt_find_common_selectors_around_adjacent_sibling_selector():
    assert_extends(
        'a + b c .c1',
        'a c .c2 {@extend .c1}',
        'a + b c .c1, a + b a c .c2, a a + b c .c2')
    assert_extends(
        'a + b c .c1',
        'a b .c2 {@extend .c1}',
        'a + b c .c1, a a + b c .c2')
    assert_extends(
        'a + b c .c1',
        'b c .c2 {@extend .c1}',
        'a + b c .c1, a + b c .c2')

def test_nested_extender_doesnt_find_common_selectors_around_sibling_selector():
    assert_extends(
        'a ~ b c .c1',
        'a c .c2 {@extend .c1}',
        'a ~ b c .c1, a ~ b a c .c2, a a ~ b c .c2')
    assert_extends(
        'a ~ b c .c1',
        'a b .c2 {@extend .c1}',
        'a ~ b c .c1, a a ~ b c .c2')
    assert_extends(
        'a ~ b c .c1',
        'b c .c2 {@extend .c1}',
        'a ~ b c .c1, a ~ b c .c2')

def test_nested_extender_doesnt_find_common_selectors_around_reference_selector():
    assert_extends(
        'a /for/ b c .c1',
        'a c .c2 {@extend .c1}',
        'a /for/ b c .c1, a /for/ b a c .c2, a a /for/ b c .c2')
    assert_extends(
        'a /for/ b c .c1',
        'a b .c2 {@extend .c1}',
        'a /for/ b c .c1, a a /for/ b c .c2')
    assert_extends(
        'a /for/ b c .c1',
        'b c .c2 {@extend .c1}',
        'a /for/ b c .c1, a /for/ b c .c2')

def test_nested_extender_with_early_child_selectors_doesnt_subseq_them():
    assert_extends('.bip > .bap .foo', '.grip > .bap .bar {@extend .foo}',
      '.bip > .bap .foo, .bip > .bap .grip > .bap .bar, .grip > .bap .bip > .bap .bar')
    assert_extends('.bap > .bip .foo', '.bap > .grip .bar {@extend .foo}',
      '.bap > .bip .foo, .bap > .bip .bap > .grip .bar, .bap > .grip .bap > .bip .bar')

def test_nested_extender_with_child_selector_unifies():
    assert_extends(
        '.baz.foo',
        'foo > bar {@extend .foo}',
        '.baz.foo, foo > bar.baz')

    assert_rendering('''\
.baz > {
  .foo {a: b}
  .bar {@extend .foo}
}
''', '''\
.baz > .foo, .baz > .bar {
  a: b; }
''')

    assert_rendering('''\
.foo {
  .bar {a: b}
  > .baz {@extend .bar}
}
''', '''\
.foo .bar, .foo > .baz {
  a: b; }
''')

def test_nested_extender_with_early_child_selectors_doesnt_subseq_them():
    assert_rendering('''\
.foo {
  .bar {a: b}
  .bip > .baz {@extend .bar}
}
''', '''\
.foo .bar, .foo .bip > .baz {
  a: b; }
''')

    assert_rendering('''\
.foo {
  .bip .bar {a: b}
  > .baz {@extend .bar}
}
''', '''\
.foo .bip .bar, .foo .bip .foo > .baz {
  a: b; }
''')

    assert_extends(
        '.foo > .bar',
        '.bip + .baz {@extend .bar}',
        '.foo > .bar, .foo > .bip + .baz')
    assert_extends(
        '.foo + .bar',
        '.bip > .baz {@extend .bar}',
        '.foo + .bar, .bip > .foo + .baz')
    assert_extends(
        '.foo > .bar',
        '.bip > .baz {@extend .bar}',
        '.foo > .bar, .bip.foo > .baz')

def test_nested_extender_with_trailing_child_selector():
    with pytest.raises(SyntaxError):
        # "bar > can't extend: invalid selector"
        render("bar > {@extend .baz}")

def test_nested_extender_with_sibling_selector():
    assert_extends(
        '.baz .foo',
        'foo + bar {@extend .foo}',
        '.baz .foo, .baz foo + bar')

def test_nested_extender_with_hacky_selector():
    assert_extends('.baz .foo', 'foo + > > + bar {@extend .foo}',
      '.baz .foo, .baz foo + > > + bar, foo .baz + > > + bar')
    assert_extends(
        '.baz .foo',
        '> > bar {@extend .foo}',
        '.baz .foo, > > .baz bar')

def test_nested_extender_merges_with_same_selector():
    assert_rendering('''\
.foo {
  .bar {a: b}
  .baz {@extend .bar} }
''', '''\
.foo .bar, .foo .baz {
  a: b; }
''')

def test_nested_extender_with_child_selector_merges_with_same_selector():
    assert_extends('.foo > .bar .baz', '.foo > .bar .bang {@extend .baz}',
      '.foo > .bar .baz, .foo > .bar .bang')

  # Combinator Unification

def test_combinator_unification_for_hacky_combinators():
    assert_extends(
        '.a > + x',
        '.b y {@extend x}',
        '.a > + x, .a .b > + y, .b .a > + y')
    assert_extends(
        '.a x',
        '.b > + y {@extend x}',
        '.a x, .a .b > + y, .b .a > + y')
    assert_extends(
        '.a > + x',
        '.b > + y {@extend x}',
        '.a > + x, .a .b > + y, .b .a > + y')
    assert_extends(
        '.a ~ > + x',
        '.b > + y {@extend x}',
        '.a ~ > + x, .a .b ~ > + y, .b .a ~ > + y')
    assert_extends(
        '.a + > x',
        '.b > + y {@extend x}',
        '.a + > x')
    assert_extends(
        '.a + > x',
        '.b > + y {@extend x}',
        '.a + > x')
    assert_extends(
        '.a ~ > + .b > x',
        '.c > + .d > y {@extend x}',
        '.a ~ > + .b > x, .a .c ~ > + .d.b > y, .c .a ~ > + .d.b > y')

def test_combinator_unification_double_tilde():
    assert_extends(
        '.a.b ~ x',
        '.a ~ y {@extend x}',
        '.a.b ~ x, .a.b ~ y')
    assert_extends(
        '.a ~ x',
        '.a.b ~ y {@extend x}',
        '.a ~ x, .a.b ~ y')
    assert_extends(
        '.a ~ x',
        '.b ~ y {@extend x}',
        '.a ~ x, .a ~ .b ~ y, .b ~ .a ~ y, .b.a ~ y')
    assert_extends(
        'a.a ~ x',
        'b.b ~ y {@extend x}',
        'a.a ~ x, a.a ~ b.b ~ y, b.b ~ a.a ~ y')

def test_combinator_unification_tilde_plus():
    assert_extends(
        '.a.b + x',
        '.a ~ y {@extend x}',
        '.a.b + x, .a.b + y')
    assert_extends(
        '.a + x',
        '.a.b ~ y {@extend x}',
        '.a + x, .a.b ~ .a + y, .a.b + y')
    assert_extends(
        '.a + x',
        '.b ~ y {@extend x}',
        '.a + x, .b ~ .a + y, .b.a + y')
    assert_extends(
        'a.a + x',
        'b.b ~ y {@extend x}',
        'a.a + x, b.b ~ a.a + y')
    assert_extends(
        '.a.b ~ x',
        '.a + y {@extend x}',
        '.a.b ~ x, .a.b ~ .a + y, .a.b + y')
    assert_extends(
        '.a ~ x',
        '.a.b + y {@extend x}',
        '.a ~ x, .a.b + y')
    assert_extends(
        '.a ~ x',
        '.b + y {@extend x}',
        '.a ~ x, .a ~ .b + y, .a.b + y')
    assert_extends(
        'a.a ~ x',
        'b.b + y {@extend x}',
        'a.a ~ x, a.a ~ b.b + y')

def test_combinator_unification_angle_sibling():
    assert_extends(
        '.a > x',
        '.b ~ y {@extend x}',
        '.a > x, .a > .b ~ y')
    assert_extends(
        '.a > x',
        '.b + y {@extend x}',
        '.a > x, .a > .b + y')
    assert_extends(
        '.a ~ x',
        '.b > y {@extend x}',
        '.a ~ x, .b > .a ~ y')
    assert_extends(
        '.a + x',
        '.b > y {@extend x}',
        '.a + x, .b > .a + y')

def test_combinator_unification_double_angle():
    assert_extends(
        '.a.b > x',
        '.b > y {@extend x}',
        '.a.b > x, .b.a > y')
    assert_extends(
        '.a > x',
        '.a.b > y {@extend x}',
        '.a > x, .a.b > y')
    assert_extends(
        '.a > x',
        '.b > y {@extend x}',
        '.a > x, .b.a > y')
    assert_extends(
        'a.a > x',
        'b.b > y {@extend x}',
        'a.a > x')

def test_combinator_unification_double_plus():
    assert_extends(
        '.a.b + x',
        '.b + y {@extend x}',
        '.a.b + x, .b.a + y')
    assert_extends(
        '.a + x',
        '.a.b + y {@extend x}',
        '.a + x, .a.b + y')
    assert_extends(
        '.a + x',
        '.b + y {@extend x}',
        '.a + x, .b.a + y')
    assert_extends(
        'a.a + x',
        'b.b + y {@extend x}',
        'a.a + x')

def test_combinator_unification_angle_space():
    assert_extends(
        '.a.b > x',
        '.a y {@extend x}',
        '.a.b > x, .a.b > y')
    assert_extends(
        '.a > x',
        '.a.b y {@extend x}',
        '.a > x, .a.b .a > y')
    assert_extends(
        '.a > x',
        '.b y {@extend x}',
        '.a > x, .b .a > y')
    assert_extends(
        '.a.b x',
        '.a > y {@extend x}',
        '.a.b x, .a.b .a > y')
    assert_extends(
        '.a x',
        '.a.b > y {@extend x}',
        '.a x, .a.b > y')
    assert_extends(
        '.a x',
        '.b > y {@extend x}',
        '.a x, .a .b > y')

def test_combinator_unification_plus_space():
    assert_extends(
        '.a.b + x',
        '.a y {@extend x}',
        '.a.b + x, .a .a.b + y')
    assert_extends(
        '.a + x',
        '.a.b y {@extend x}',
        '.a + x, .a.b .a + y')
    assert_extends(
        '.a + x',
        '.b y {@extend x}',
        '.a + x, .b .a + y')
    assert_extends(
        '.a.b x',
        '.a + y {@extend x}',
        '.a.b x, .a.b .a + y')
    assert_extends(
        '.a x',
        '.a.b + y {@extend x}',
        '.a x, .a .a.b + y')
    assert_extends(
        '.a x',
        '.b + y {@extend x}',
        '.a x, .a .b + y')

def test_combinator_unification_nested():
    assert_extends(
        '.a > .b + x',
        '.c > .d + y {@extend x}',
        '.a > .b + x, .c.a > .d.b + y')
    assert_extends(
        '.a > .b + x',
        '.c > y {@extend x}',
        '.a > .b + x, .c.a > .b + y')

def test_combinator_unification_with_newlines():
    assert_rendering('''\
.a >
.b
+ x {a: b}
.c
> .d +
y {@extend x}
''', '''\
.a >
.b
+ x, .c.a > .d.b + y {
  a: b; }
''')

  # Loops

def test_extend_self_loop():
    assert_rendering('''\
.foo {a: b; @extend .foo}
''', '''\
.foo {
  a: b; }
''')

def test_basic_extend_loop():
    assert_rendering('''\
.foo {a: b; @extend .bar}
.bar {c: d; @extend .foo}
''', '''\
.bar, .foo {
  a: b; }

.foo, .bar {
  c: d; }
''')

def test_three_level_extend_loop():
    assert_rendering('''\
.foo {a: b; @extend .bar}
.bar {c: d; @extend .baz}
.baz {e: f; @extend .foo}
''', '''\
.baz, .bar, .foo {
  a: b; }

.foo, .baz, .bar {
  c: d; }

.bar, .foo, .baz {
  e: f; }
''')

def test_nested_extend_loop():
    assert_rendering('''\
.bar {
  a: b;
  .foo {c: d; @extend .bar}
}
''', '''\
.bar, .bar .foo {
  a: b; }
  .bar .foo {
    c: d; }
''')

def test_multiple_extender_merges_with_superset_selector():
    assert_rendering('''\
.foo {@extend .bar; @extend .baz}
a.bar.baz {a: b}
''', '''\
a.bar.baz, a.foo {
  a: b; }
''')

def test_control_flow_if():
    assert_rendering('''\
.true  { color: green; }
.false { color: red;   }
.also-true {
  @if true { @extend .true;  }
  @else    { @extend .false; }
}
.also-false {
  @if false { @extend .true;  }
  @else     { @extend .false; }
}
''', '''\
.true, .also-true {
  color: green; }

.false, .also-false {
  color: red; }
''')

def test_control_flow_for():
    assert_rendering('''\
.base-0  { color: green; }
.base-1  { display: block; }
.base-2  { border: 1px solid blue; }
.added {
  @for $i from 0 to 3 {
    @extend .base-\#{$i};
  }
}
''', '''\
.base-0, .added {
  color: green; }

.base-1, .added {
  display: block; }

.base-2, .added {
  border: 1px solid blue; }
''')

def test_control_flow_while():
    assert_rendering('''\
.base-0  { color: green; }
.base-1  { display: block; }
.base-2  { border: 1px solid blue; }
.added {
  $i : 0;
  @while $i < 3 {
    @extend .base-\#{$i};
    $i : $i + 1;
  }
}
''', '''\
.base-0, .added {
  color: green; }

.base-1, .added {
  display: block; }

.base-2, .added {
  border: 1px solid blue; }
''')

def test_basic_placeholder_selector():
    assert_extends(
        '%foo',
        '.bar {@extend %foo}',
        '.bar')

def test_unused_placeholder_selector():
    assert_rendering('''\
%foo {color: blue}
%bar {color: red}
.baz {@extend %foo}
''', '''\
.baz {
  color: blue; }
''')

def test_placeholder_descendant_selector():
    assert_extends(
        '#context %foo a',
        '.bar {@extend %foo}',
        '#context .bar a')

def test_semi_placeholder_selector():
    assert_rendering('''\
#context %foo, .bar .baz {color: blue}
''', '''\
.bar .baz {
  color: blue; }
''')

def test_placeholder_selector_with_multiple_extenders():
    assert_rendering('''\
%foo {color: blue}
.bar {@extend %foo}
.baz {@extend %foo}
''', '''\
.bar, .baz {
  color: blue; }
''')

def test_placeholder_selector_as_modifier():
    # XXX assert_extend_doesnt_match('div', '%foo', :failed_to_unify, 3) do
    assert_rendering('''\
a%foo.baz {color: blue}
.bar {@extend %foo}
div {@extend %foo}
''', '''\
a.baz.bar {
  color: blue; }
''')

def test_placeholder_interpolation():
    assert_rendering('''\
$foo: foo;

%\#{$foo} {color: blue}
.bar {@extend %foo}
''', '''\
.bar {
  color: blue; }
''')

def test_media_in_placeholder_selector():
    assert_rendering('''\
%foo {bar {@media screen {a: b}}}
.baz {c: d}
''', '''\
.baz {
  c: d; }
''')

"""
def test_extend_out_of_media():
    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
DEPRECATION WARNING on line 3 of test_extend_out_of_media_inline.scss:
  @extending an outer selector from within @media is deprecated.
  You may only @extend selectors within the same directive.
  This will be an error in Sass 3.3.
  It can only work once @extend is supported natively in the browser.
WARN
.foo {
  a: b; }
CSS
.foo {a: b}
@media screen {
  .bar {@extend .foo}
}
SCSS

def test_extend_out_of_unknown_directive():
    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
DEPRECATION WARNING on line 3 of test_extend_out_of_unknown_directive_inline.scss:
  @extending an outer selector from within @flooblehoof is deprecated.
  You may only @extend selectors within the same directive.
  This will be an error in Sass 3.3.
  It can only work once @extend is supported natively in the browser.
WARN
.foo {
  a: b; }

@flooblehoof {}
CSS
.foo {a: b}
@flooblehoof {
  .bar {@extend .foo}
}
SCSS

def test_extend_out_of_nested_directives():
    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
DEPRECATION WARNING on line 4 of test_extend_out_of_nested_directives_inline.scss:
  @extending an outer selector from within @flooblehoof is deprecated.
  You may only @extend selectors within the same directive.
  This will be an error in Sass 3.3.
  It can only work once @extend is supported natively in the browser.
WARN
@media screen {
  .foo {
    a: b; }

  @flooblehoof {} }
CSS
@media screen {
  .foo {a: b}
  @flooblehoof {
    .bar {@extend .foo}
  }
}
SCSS
"""

def test_extend_within_media():
    assert_rendering('''\
@media screen {
  .foo {a: b}
  .bar {@extend .foo}
}
''', '''\
@media screen {
  .foo, .bar {
    a: b; } }
''')

def test_extend_within_unknown_directive():
    assert_rendering('''\
@flooblehoof {
  .foo {a: b}
  .bar {@extend .foo}
}
''', '''\
@flooblehoof {
  .foo, .bar {
    a: b; } }
''')

def test_extend_within_nested_directives():
    assert_rendering('''\
@media screen {
  @flooblehoof {
    .foo {a: b}
    .bar {@extend .foo}
  }
}
''', '''\
@media screen {
  @flooblehoof {
    .foo, .bar {
      a: b; } } }
''')

def test_extend_within_disparate_media():
    assert_rendering('''\
@media screen {.foo {a: b}}
@media screen {.bar {@extend .foo}}
''', '''\
@media screen {
  .foo, .bar {
    a: b; } }
''')

def test_extend_within_disparate_unknown_directive():
    assert_rendering('''\
@flooblehoof {.foo {a: b}}
@flooblehoof {.bar {@extend .foo}}
''', '''\
@flooblehoof {
  .foo, .bar {
    a: b; } }

@flooblehoof {}
''')

def test_extend_within_disparate_nested_directives():
    assert_rendering('''\
@media screen {@flooblehoof {.foo {a: b}}}
@media screen {@flooblehoof {.bar {@extend .foo}}}
''', '''\
@media screen {
  @flooblehoof {
    .foo, .bar {
      a: b; } } }
@media screen {
  @flooblehoof {} }
''')

"""
def test_extend_within_and_without_media():
    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
DEPRECATION WARNING on line 4 of test_extend_within_and_without_media_inline.scss:
  @extending an outer selector from within @media is deprecated.
  You may only @extend selectors within the same directive.
  This will be an error in Sass 3.3.
  It can only work once @extend is supported natively in the browser.
WARN
.foo {
  a: b; }

@media screen {
  .foo, .bar {
    c: d; } }
CSS
.foo {a: b}
@media screen {
  .foo {c: d}
  .bar {@extend .foo}
}
SCSS

def test_extend_within_and_without_unknown_directive():
    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
DEPRECATION WARNING on line 4 of test_extend_within_and_without_unknown_directive_inline.scss:
  @extending an outer selector from within @flooblehoof is deprecated.
  You may only @extend selectors within the same directive.
  This will be an error in Sass 3.3.
  It can only work once @extend is supported natively in the browser.
WARN
.foo {
  a: b; }

@flooblehoof {
  .foo, .bar {
    c: d; } }
CSS
.foo {a: b}
@flooblehoof {
  .foo {c: d}
  .bar {@extend .foo}
}
SCSS

def test_extend_within_and_without_nested_directives():
    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
DEPRECATION WARNING on line 5 of test_extend_within_and_without_nested_directives_inline.scss:
  @extending an outer selector from within @flooblehoof is deprecated.
  You may only @extend selectors within the same directive.
  This will be an error in Sass 3.3.
  It can only work once @extend is supported natively in the browser.
WARN
@media screen {
  .foo {
    a: b; }

  @flooblehoof {
    .foo, .bar {
      c: d; } } }
CSS
@media screen {
  .foo {a: b}
  @flooblehoof {
    .foo {c: d}
    .bar {@extend .foo}
  }
}
SCSS
"""

def test_extend_with_subject_transfers_subject_to_extender():
    assert_rendering('''\
foo bar! baz {a: b}
.bip .bap {@extend bar}
''', '''\
foo bar! baz, foo .bip .bap! baz, .bip foo .bap! baz {
  a: b; }
''')

    assert_rendering('''\
foo.x bar.y! baz.z {a: b}
.bip .bap {@extend .y}
''', '''\
foo.x bar.y! baz.z, foo.x .bip bar.bap! baz.z, .bip foo.x bar.bap! baz.z {
  a: b; }
''')

def test_extend_with_subject_retains_subject_on_target():
    assert_rendering('''\
.foo! .bar {a: b}
.bip .bap {@extend .bar}
''', '''\
.foo! .bar, .foo! .bip .bap, .bip .foo! .bap {
  a: b; }
''')

def test_extend_with_subject_transfers_subject_to_target():
    assert_rendering('''\
a.foo .bar {a: b}
.bip .bap! {@extend .foo}
''', '''\
a.foo .bar, .bip a.bap! .bar {
  a: b; }
''')

def test_extend_with_subject_retains_subject_on_extender():
    assert_rendering('''\
.foo .bar {a: b}
.bip! .bap {@extend .bar}
''', '''\
.foo .bar, .foo .bip! .bap, .bip! .foo .bap {
  a: b; }
''')

def test_extend_with_subject_fails_with_conflicting_subject():
    assert_rendering('''\
x! .bar {a: b}
y! .bap {@extend .bar}
''', '''\
x! .bar {
  a: b; }
''')

"""
def test_extend_warns_when_extendee_doesnt_exist():
    assert_warning(<<WARN) {assert_equal("", render(<<SCSS))}
WARNING on line 1 of test_extend_warns_when_extendee_doesnt_exist_inline.scss: ".foo" failed to @extend ".bar".
  The selector ".bar" was not found.
  This will be an error in future releases of Sass.
  Use "@extend .bar !optional" if the extend should be able to fail.
WARN
.foo {@extend .bar}
SCSS

def test_extend_warns_when_extension_fails():
    assert_warning(<<WARN) {assert_equal(<<CSS, render(<<SCSS))}
WARNING on line 2 of test_extend_warns_when_extension_fails_inline.scss: "b.foo" failed to @extend ".bar".
  No selectors matching ".bar" could be unified with "b.foo".
  This will be an error in future releases of Sass.
  Use "@extend .bar !optional" if the extend should be able to fail.
WARN
a.bar {
  a: b; }
CSS
a.bar {a: b}
b.foo {@extend .bar}
SCSS

def test_extend_does_not_warn_when_one_extension_fails_but_others_dont():
    assert_no_warning {assert_equal(<<CSS, render(<<SCSS))}
a.bar {
  a: b; }

.bar, b.foo {
  c: d; }
CSS
a.bar {a: b}
.bar {c: d}
b.foo {@extend .bar}
SCSS

def test_extend_does_not_warn_when_one_extension_fails_but_others_dont():
    assert_no_warning {assert_equal(<<CSS, render(<<SCSS))}
a.bar {
  a: b; }

.bar, b.foo {
  c: d; }
CSS
a.bar {a: b}
.bar {c: d}
b.foo {@extend .bar}
SCSS

def test_optional_extend_does_not_warn_when_extendee_doesnt_exist():
    assert_no_warning {assert_equal("", render(<<SCSS))}
.foo {@extend .bar !optional}
SCSS

def test_optional_extend_does_not_warn_when_extension_fails():
    assert_no_warning {assert_equal(<<CSS, render(<<SCSS))}
a.bar {
  a: b; }
CSS
a.bar {a: b}
b.foo {@extend .bar !optional}
SCSS
"""


### Regression Tests

def test_nested_extend_specificity():
    assert_rendering('''\
%foo {a: b}

a {
  :b {@extend %foo}
  :b:c {@extend %foo}
}
''', '''\
a :b, a :b:c {
  a: b; }
''')

def test_nested_double_extend_optimization():
    assert_rendering('''\
%foo %bar {
  a: b;
}

.parent1 {
  @extend %foo;

  .child {
    @extend %bar;
  }
}

.parent2 {
  @extend %foo;
}
''', '''\
.parent1 .child {
  a: b; }
''')

def test_extend_in_double_nested_media_query():
    assert_rendering('''\
@media all {
  @media (orientation: landscape) {
    %foo {color: blue}
    .bar {@extend %foo}
  }
}
''', '''\
@media all and (orientation: landscape) {
  .bar {
    color: blue; } }
''')

"""
def test_partially_failed_extend():
    assert_no_warning {assert_equal(<<CSS, render(<<SCSS))}
.rc, test {
  color: white; }

.prices span.pill span.rc {
  color: red; }
CSS
test { @extend .rc; }
.rc {color: white;}
.prices span.pill span.rc {color: red;}
SCSS
"""

def test_newline_near_combinator():
    assert_rendering('''\
.a +
.b x {a: b}
.c y {@extend x}
''', '''\
.a +
.b x, .a +
.b .c y, .c .a +
.b y {
  a: b; }
''')

def test_duplicated_selector_with_newlines():
    assert_rendering('''\
.example-1-1,
.example-1-2,
.example-1-3 {
  a: b;
}

.my-page-1 .my-module-1-1 {@extend .example-1-2}
''', '''\
.example-1-1,
.example-1-2,
.my-page-1 .my-module-1-1,
.example-1-3 {
  a: b; }
''')

def test_nested_selector_with_child_selector_hack_extendee():
    assert_extends(
        '> .foo',
        'foo bar {@extend .foo}',
        '> .foo, > foo bar')

def test_nested_selector_with_child_selector_hack_extender():
    assert_extends(
        '.foo .bar',
        '> foo bar {@extend .bar}',
        '.foo .bar, > .foo foo bar, > foo .foo bar')

def test_nested_selector_with_child_selector_hack_extender_and_extendee():
    assert_extends(
        '> .foo',
        '> foo bar {@extend .foo}',
        '> .foo, > foo bar')

def test_nested_selector_with_child_selector_hack_extender_and_sibling_selector_extendee():
    assert_extends(
        '~ .foo',
        '> foo bar {@extend .foo}',
        '~ .foo')

def test_nested_selector_with_child_selector_hack_extender_and_extendee_and_newline():
    assert_rendering('''\
> .foo {a: b}
flip,
> foo bar {@extend .foo}
''', '''\
> .foo, > flip,
> foo bar {
  a: b; }
''')

def test_extended_parent_and_child_redundancy_elimination():
    assert_rendering('''\
a {
  b {a: b}
  c {@extend b}
}
d {@extend a}
''', '''\
a b, d b, a c, d c {
  a: b; }
''')

def test_extend_redundancy_elimination_when_it_would_reduce_specificity():
    assert_extends(
        'a',
        'a.foo {@extend a}',
        'a, a.foo')

def test_extend_redundancy_elimination_when_it_would_preserve_specificity():
    assert_extends(
        '.bar a',
        'a.foo {@extend a}',
        '.bar a')

def test_extend_redundancy_elimination_never_eliminates_base_selector():
    assert_extends(
        'a.foo',
        '.foo {@extend a}',
        'a.foo, .foo')

def test_extend_cross_branch_redundancy_elimination():
    assert_rendering('''\
%x c %y {a: b}
a, b {@extend %x}
a d {@extend %y}
''', '''\
a c d, b c a d {
  a: b; }
''')

    assert_rendering('''\
e %z {a: b}
%x c %y {@extend %z}
a, b {@extend %x}
a d {@extend %y}
''', '''\
e a c d, a c e d, e b c a d, b c a e d {
  a: b; }
''')


"""
def assert_extend_doesnt_match(extender, target, reason, line, syntax = :scss):
    warn = "\"#{extender}\" failed to @extend \"#{target}\"."
    reason =
      if reason == :not_found
        "The selector \"#{target}\" was not found."
      else
        "No selectors matching \"#{target}\" could be unified with \"#{extender}\"."

    assert_warning(<<WARNING) {yield}
WARNING on line #{line} of #{filename_for_test syntax}: #{warn}
  #{reason}
  This will be an error in future releases of Sass.
  Use "@extend #{target} !optional" if the extend should be able to fail.
WARNING
"""

def assert_unification(selector, extension, unified):
    # Do some trickery so the first law of extend doesn't get in our way.
    assert_extends(
      "%-a {0}".format(selector),
      extension + " -a {@extend %-a}",
      ', '.join('-a ' + s for s in unified.split(', ')))

def assert_extends(selector, extension, result):
    assert_rendering(
        "{0} {{a: b}}\n{1}\n".format(selector, extension),
        "{0} {{\n  a: b; }}\n".format(result))



"""Utilities for working with gradients.  Inspired by Compass, but not quite
the same.
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import base64
import logging

import six

from . import CompassExtension
from .helpers import opposite_position, position
from scss.types import Color, List, Number, String
from scss.util import escape, split_params, to_float, to_str

log = logging.getLogger(__name__)
ns = CompassExtension.namespace


def _is_color(value):
    # currentColor is not a Sass color value, but /is/ a CSS color value
    return isinstance(value, Color) or value == String('currentColor')


def __color_stops(percentages, *args):
    if len(args) == 1:
        if isinstance(args[0], (list, tuple, List)):
            return list(args[0])
        elif isinstance(args[0], (String, six.string_types)):
            color_stops = []
            colors = split_params(getattr(args[0], 'value', args[0]))
            for color in colors:
                color = color.strip()
                if color.startswith('color-stop('):
                    s, c = split_params(color[11:].rstrip(')'))
                    s = s.strip()
                    c = c.strip()
                else:
                    c, s = color.split()
                color_stops.append((to_float(s), c))
            return color_stops

    colors = []
    stops = []
    prev_color = False
    for c in args:
        for c in List.from_maybe(c):
            if _is_color(c):
                if prev_color:
                    stops.append(None)
                colors.append(c)
                prev_color = True
            elif isinstance(c, Number):
                stops.append(c)
                prev_color = False

    if prev_color:
        stops.append(None)
    stops = stops[:len(colors)]
    if stops[0] is None:
        stops[0] = Number(0, '%')
    if stops[-1] is None:
        stops[-1] = Number(100, '%')

    maxable_stops = [s for s in stops if s and not s.is_simple_unit('%')]
    if maxable_stops:
        max_stops = max(maxable_stops)
    else:
        max_stops = None

    stops = [_s / max_stops if _s and not _s.is_simple_unit('%') else _s for _s in stops]

    init = 0
    start = None
    for i, s in enumerate(stops + [1.0]):
        if s is None:
            if start is None:
                start = i
            end = i
        else:
            final = s
            if start is not None:
                stride = (final - init) / Number(end - start + 1 + (1 if i < len(stops) else 0))
                for j in range(start, end + 1):
                    stops[j] = init + stride * Number(j - start + 1)
            init = final
            start = None

    if not max_stops or percentages:
        pass
    else:
        stops = [s if s.is_simple_unit('%') else s * max_stops for s in stops]

    return List(List(pair) for pair in zip(stops, colors))


def _render_standard_color_stops(color_stops):
    pairs = []
    for i, (stop, color) in enumerate(color_stops):
        if ((i == 0 and stop == Number(0, '%')) or
                (i == len(color_stops) - 1 and stop == Number(100, '%'))):
            pairs.append(color)
        else:
            pairs.append(List([color, stop], use_comma=False))

    return List(pairs, use_comma=True)


@ns.declare
def grad_color_stops(*args):
    args = List.from_maybe_starargs(args)
    color_stops = __color_stops(True, *args)
    ret = ', '.join(['color-stop(%s, %s)' % (s.render(), c.render()) for s, c in color_stops])
    return String.unquoted(ret)


def __grad_end_position(radial, color_stops):
    return __grad_position(-1, 100, radial, color_stops)


@ns.declare
def grad_point(*p):
    pos = set()
    hrz = vrt = Number(0.5, '%')
    for _p in p:
        pos.update(String.unquoted(_p).value.split())
    if 'left' in pos:
        hrz = Number(0, '%')
    elif 'right' in pos:
        hrz = Number(1, '%')
    if 'top' in pos:
        vrt = Number(0, '%')
    elif 'bottom' in pos:
        vrt = Number(1, '%')
    return List([v for v in (hrz, vrt) if v is not None])


def __grad_position(index, default, radial, color_stops):
    try:
        stops = Number(color_stops[index][0])
        if radial and not stops.is_simple_unit('px') and (index == 0 or index == -1 or index == len(color_stops) - 1):
            log.warn("Webkit only supports pixels for the start and end stops for radial gradients. Got %s", stops)
    except IndexError:
        stops = Number(default)
    return stops


@ns.declare
def grad_end_position(*color_stops):
    color_stops = __color_stops(False, *color_stops)
    return Number(__grad_end_position(False, color_stops))


@ns.declare
def color_stops(*args):
    args = List.from_maybe_starargs(args)
    color_stops = __color_stops(False, *args)
    ret = ', '.join(['%s %s' % (c.render(), s.render()) for s, c in color_stops])
    return String.unquoted(ret)


@ns.declare
def color_stops_in_percentages(*args):
    args = List.from_maybe_starargs(args)
    color_stops = __color_stops(True, *args)
    ret = ', '.join(['%s %s' % (c.render(), s.render()) for s, c in color_stops])
    return String.unquoted(ret)


def _get_gradient_position_and_angle(args):
    for arg in args:
        ret = None
        skip = False
        for a in arg:
            if _is_color(a):
                skip = True
                break
            elif isinstance(a, Number):
                ret = arg
        if skip:
            continue
        if ret is not None:
            return ret
        for seek in (
            'center',
            'top', 'bottom',
            'left', 'right',
        ):
            if String(seek) in arg:
                return arg
    return None


def _get_gradient_shape_and_size(args):
    for arg in args:
        for seek in (
            'circle', 'ellipse',
            'closest-side', 'closest-corner',
            'farthest-side', 'farthest-corner',
            'contain', 'cover',
        ):
            if String(seek) in arg:
                return arg
    return None


def _get_gradient_color_stops(args):
    color_stops = []
    for arg in args:
        for a in List.from_maybe(arg):
            if _is_color(a):
                color_stops.append(arg)
                break
    return color_stops or None


# TODO these functions need to be
# 1. well-defined
# 2. guaranteed to never wreck css3 syntax
# 3. updated to whatever current compass does
# 4. fixed to use a custom type instead of monkeypatching


@ns.declare
def radial_gradient(*args):
    args = List.from_maybe_starargs(args)

    try:
        # Do a rough check for standard syntax first -- `shape at position`
        at_position = list(args[0]).index(String('at'))
    except (IndexError, ValueError):
        shape_and_size = _get_gradient_shape_and_size(args)
        position_and_angle = _get_gradient_position_and_angle(args)
    else:
        shape_and_size = List.maybe_new(args[0][:at_position])
        position_and_angle = List.maybe_new(args[0][at_position + 1:])

    color_stops = _get_gradient_color_stops(args)
    if color_stops is None:
        raise Exception('No color stops provided to radial-gradient function')
    color_stops = __color_stops(False, *color_stops)

    if position_and_angle:
        rendered_position = position(position_and_angle)
    else:
        rendered_position = None
    rendered_color_stops = _render_standard_color_stops(color_stops)

    args = []
    if shape_and_size and rendered_position:
        args.append(List([shape_and_size, String.unquoted('at'), rendered_position], use_comma=False))
    elif rendered_position:
        args.append(rendered_position)
    elif shape_and_size:
        args.append(shape_and_size)
    args.extend(rendered_color_stops)

    legacy_args = []
    if rendered_position:
        legacy_args.append(rendered_position)
    if shape_and_size:
        legacy_args.append(shape_and_size)
    legacy_args.extend(rendered_color_stops)

    ret = String.unquoted(
        'radial-gradient(' + ', '.join(a.render() for a in args) + ')')

    legacy_ret = 'radial-gradient(' + ', '.join(a.render() for a in legacy_args) + ')'

    def to__css2():
        return String.unquoted('')
    ret.to__css2 = to__css2

    def to__moz():
        return String.unquoted('-moz-' + legacy_ret)
    ret.to__moz = to__moz

    def to__pie():
        log.warn("PIE does not support radial-gradient.")
        return String.unquoted('-pie-radial-gradient(unsupported)')
    ret.to__pie = to__pie

    def to__webkit():
        return String.unquoted('-webkit-' + legacy_ret)
    ret.to__webkit = to__webkit

    def to__owg():
        args = [
            'radial',
            grad_point(*position_and_angle) if position_and_angle is not None else 'center',
            '0',
            grad_point(*position_and_angle) if position_and_angle is not None else 'center',
            __grad_end_position(True, color_stops),
        ]
        args.extend('color-stop(%s, %s)' % (s.render(), c.render()) for s, c in color_stops)
        ret = '-webkit-gradient(' + ', '.join(to_str(a) for a in args or [] if a is not None) + ')'
        return String.unquoted(ret)
    ret.to__owg = to__owg

    def to__svg():
        return radial_svg_gradient(*(list(color_stops) + list(position_and_angle or [String('center')])))
    ret.to__svg = to__svg

    return ret


@ns.declare
def linear_gradient(*args):
    args = List.from_maybe_starargs(args)

    position_and_angle = _get_gradient_position_and_angle(args)
    color_stops = _get_gradient_color_stops(args)
    if color_stops is None:
        raise Exception('No color stops provided to linear-gradient function')
    color_stops = __color_stops(False, *color_stops)

    args = [
        position(position_and_angle) if position_and_angle is not None else None,
    ]
    args.extend(_render_standard_color_stops(color_stops))

    to__s = 'linear-gradient(' + ', '.join(to_str(a) for a in args or [] if a is not None) + ')'
    ret = String.unquoted(to__s)

    def to__css2():
        return String.unquoted('')
    ret.to__css2 = to__css2

    def to__moz():
        return String.unquoted('-moz-' + to__s)
    ret.to__moz = to__moz

    def to__pie():
        return String.unquoted('-pie-' + to__s)
    ret.to__pie = to__pie

    def to__ms():
        return String.unquoted('-ms-' + to__s)
    ret.to__ms = to__ms

    def to__o():
        return String.unquoted('-o-' + to__s)
    ret.to__o = to__o

    def to__webkit():
        return String.unquoted('-webkit-' + to__s)
    ret.to__webkit = to__webkit

    def to__owg():
        args = [
            'linear',
            position(position_and_angle or None),
            opposite_position(position_and_angle or None),
        ]
        args.extend('color-stop(%s, %s)' % (s.render(), c.render()) for s, c in color_stops)
        ret = '-webkit-gradient(' + ', '.join(to_str(a) for a in args if a is not None) + ')'
        return String.unquoted(ret)
    ret.to__owg = to__owg

    def to__svg():
        return linear_svg_gradient(color_stops, position_and_angle or 'top')
    ret.to__svg = to__svg

    return ret


@ns.declare
def radial_svg_gradient(*args):
    args = List.from_maybe_starargs(args)
    color_stops = args
    center = None
    if isinstance(args[-1], (String, Number)):
        center = args[-1]
        color_stops = args[:-1]
    color_stops = __color_stops(False, *color_stops)
    cx, cy = grad_point(center)
    r = __grad_end_position(True, color_stops)
    svg = __radial_svg(color_stops, cx, cy, r)
    url = 'data:' + 'image/svg+xml' + ';base64,' + base64.b64encode(svg)
    inline = 'url("%s")' % escape(url)
    return String.unquoted(inline)


@ns.declare
def linear_svg_gradient(*args):
    args = List.from_maybe_starargs(args)
    color_stops = args
    start = None
    if isinstance(args[-1], (String, Number)):
        start = args[-1]
        color_stops = args[:-1]
    color_stops = __color_stops(False, *color_stops)
    x1, y1 = grad_point(start)
    x2, y2 = grad_point(opposite_position(start))
    svg = _linear_svg(color_stops, x1, y1, x2, y2)
    url = 'data:' + 'image/svg+xml' + ';base64,' + base64.b64encode(svg)
    inline = 'url("%s")' % escape(url)
    return String.unquoted(inline)


def __color_stops_svg(color_stops):
    ret = ''.join('<stop offset="%s" stop-color="%s"/>' % (to_str(s), c) for s, c in color_stops)
    return ret


def __svg_template(gradient):
    ret = '<?xml version="1.0" encoding="utf-8"?>\
<svg version="1.1" xmlns="http://www.w3.org/2000/svg">\
<defs>%s</defs>\
<rect x="0" y="0" width="100%%" height="100%%" fill="url(#grad)" />\
</svg>' % gradient
    return ret


def _linear_svg(color_stops, x1, y1, x2, y2):
    gradient = '<linearGradient id="grad" x1="%s" y1="%s" x2="%s" y2="%s">%s</linearGradient>' % (
        to_str(Number(x1)),
        to_str(Number(y1)),
        to_str(Number(x2)),
        to_str(Number(y2)),
        __color_stops_svg(color_stops)
    )
    return __svg_template(gradient)


def __radial_svg(color_stops, cx, cy, r):
    gradient = '<radialGradient id="grad" gradientUnits="userSpaceOnUse" cx="%s" cy="%s" r="%s">%s</radialGradient>' % (
        to_str(Number(cx)),
        to_str(Number(cy)),
        to_str(Number(r)),
        __color_stops_svg(color_stops)
    )
    return __svg_template(gradient)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import six

import sys
import traceback


BROWSER_ERROR_TEMPLATE = """\
body:before {{
    content: {0};

    display: block;
    position: fixed;
    top: 0;
    left: 0;
    right: 0;

    font-size: 14px;
    margin: 1em;
    padding: 1em;
    border: 3px double red;

    white-space: pre;
    font-family: monospace;
    background: #fcebeb;
    color: black;
}}
"""


def add_error_marker(text, position, start_line=1):
    """Add a caret marking a given position in a string of input.

    Returns (new_text, caret_line).
    """
    indent = "    "
    lines = []
    caret_line = start_line
    for line in text.split("\n"):
        lines.append(indent + line)

        if 0 <= position <= len(line):
            lines.append(indent + (" " * position) + "^")
            caret_line = start_line

        position -= len(line)
        position -= 1  # for the newline
        start_line += 1

    return "\n".join(lines), caret_line


class SassBaseError(Exception):
    """Base class for all errors caused by Sass code.

    Shouldn't be raising this directly; use or create a subclass instead.
    """

    def __init__(self, rule=None):
        super(SassBaseError, self).__init__()

        self.rule_stack = []
        if rule is not None:
            self.add_rule(rule)

    def add_rule(self, rule):
        """Add a new rule to the "stack" of rules -- this is used to track,
        e.g., how a file was ultimately imported.
        """
        self.rule_stack.append(rule)

    def format_file_and_line(self, rule):
        return "line {rule.lineno} of {rule.source_file.path}".format(
            rule=rule,
        )

    def format_sass_stack(self):
        """Return a "traceback" of Sass imports."""
        if not self.rule_stack:
            return ""

        ret = ["on ", self.format_file_and_line(self.rule_stack[0]), "\n"]
        last_file = self.rule_stack[0].source_file

        # TODO this could go away if rules knew their import chains...
        # TODO this doesn't mention mixins or function calls.  really need to
        # track the call stack better.  atm we skip other calls in the same
        # file because most of them are just nesting, but they might not be!
        # TODO the line number is wrong here for @imports, because we don't
        # have access to the UnparsedBlock representing the import!
        # TODO @content is completely broken; it's basically textual inclusion
        for rule in self.rule_stack[1:]:
            if rule.source_file is not last_file:
                ret.extend((
                    "imported from ", self.format_file_and_line(rule), "\n"))
            last_file = rule.source_file

        return "".join(ret)

    def format_message(self):
        return ""

    def __str__(self):
        return "{message}\n{sass_stack}".format(
            message=self.format_message(),
            sass_stack=self.format_sass_stack(),
        )


class SassSyntaxError(SassBaseError):
    """Generic syntax error thrown by the guts of the expression parser;
    usually caught and wrapped later on.
    """
    def __init__(self, input_string, position, desired_tokens):
        super(SassSyntaxError, self).__init__()

        self.input_string = input_string
        self.position = position
        self.desired_tokens = desired_tokens

    def __str__(self):
        # TODO this doesn't show the rule stack; should inherit from SassError
        # instead?
        if self.position == 0:
            after = "Syntax error"
        else:
            after = "Syntax error after {0!r}".format(
                self.input_string[max(0, self.position - 20):self.position])

        found = "Found {0!r}".format(
            self.input_string[self.position:self.position + 10])

        if not self.desired_tokens:
            expected = "but can't figure out what that means"
        elif len(self.desired_tokens) == 1:
            expected = "but expected {0}".format(
                ''.join(self.desired_tokens))
        else:
            expected = "but expected one of {0}".format(
                ', '.join(sorted(self.desired_tokens)))

        return "{after}: {found} {expected}".format(
            after=after,
            found=found,
            expected=expected,
        )


class SassImportError(SassBaseError):
    """Error raised when unable to resolve an @import."""

    def __init__(self, bad_name, compiler, **kwargs):
        super(SassImportError, self).__init__(**kwargs)

        self.bad_name = bad_name
        self.compiler = compiler

    def format_message(self):
        return (
            "Couldn't find anything to import: {0}\n"
            "Extensions: {1}\n"
            "Search path:\n  {2}"
            .format(
                self.bad_name,
                ", ".join(repr(ext) for ext in self.compiler.extensions),
                "\n  ".join(str(path) for path in self.compiler.search_path),
            )
        )


class SassMissingDependency(SassBaseError):
    """Error raised when an optional library (e.g., PIL or fontforge) is
    missing.
    """

    def __init__(self, library, activity, **kwargs):
        super(SassMissingDependency, self).__init__(**kwargs)

        self.library = library
        self.activity = activity

    def format_message(self):
        return "{0} requires {1}, which is not installed".format(
            self.activity, self.library)


class SassError(SassBaseError):
    """Error class that wraps another exception and attempts to bolt on some
    useful context.
    """
    def __init__(self, exc, expression=None, expression_pos=None, **kwargs):
        super(SassError, self).__init__(**kwargs)

        self.exc = exc
        self.expression = expression
        self.expression_pos = expression_pos

        _, _, self.original_traceback = sys.exc_info()

    def format_prefix(self):
        """Return the general name of the error and the contents of the rule or
        property that caused the failure.  This is the initial part of the
        error message and should be error-specific.
        """
        # TODO this contains NULs and line numbers; could be much prettier
        if self.rule_stack:
            return (
                "Error parsing block:\n" +
                "    " + self.rule_stack[0].unparsed_contents + "\n"
            )
        else:
            return "Unknown error\n"

    def format_python_stack(self):
        """Return a traceback of Python frames, from where the error occurred
        to where it was first caught and wrapped.
        """
        ret = ["Traceback:\n"]
        ret.extend(traceback.format_tb(self.original_traceback))
        return "".join(ret)

    def format_original_error(self):
        """Return the typical "TypeError: blah blah" for the original wrapped
        error.
        """
        # TODO eventually we'll have sass-specific errors that will want nicer
        # "names" in browser display and stderr
        return "".join((
            type(self.exc).__name__, ": ", six.text_type(self.exc), "\n",
        ))

    def __str__(self):
        try:
            prefix = self.format_prefix()
            sass_stack = self.format_sass_stack()
            python_stack = self.format_python_stack()
            original_error = self.format_original_error()

            # TODO not very well-specified whether these parts should already
            # end in newlines, or how many
            return prefix + "\n" + sass_stack + python_stack + original_error
        except Exception:
            # "unprintable error" is not helpful
            return six.text_type(self.exc)

    def to_css(self):
        """Return a stylesheet that will show the wrapped error at the top of
        the browser window.
        """
        # TODO should this include the traceback?  any security concerns?
        prefix = self.format_prefix()
        original_error = self.format_original_error()
        sass_stack = self.format_sass_stack()

        message = prefix + "\n" + sass_stack + original_error

        # Super simple escaping: only quotes and newlines are illegal in css
        # strings
        message = message.replace('\\', '\\\\')
        message = message.replace('"', '\\"')
        # use the maximum six digits here so it doesn't eat any following
        # characters that happen to look like hex
        message = message.replace('\n', '\\00000A')

        return BROWSER_ERROR_TEMPLATE.format('"' + message + '"')


class SassParseError(SassError):
    """Error raised when parsing a Sass expression fails."""

    def format_prefix(self):
        decorated_expr, line = add_error_marker(self.expression, self.expression_pos or -1)
        return """Error parsing expression at {1}:\n{0}\n""".format(decorated_expr, self.expression_pos)


class SassEvaluationError(SassError):
    """Error raised when evaluating a parsed expression fails."""

    def format_prefix(self):
        # TODO boy this is getting repeated a lot
        # TODO would be nice for the AST to have position information
        # TODO might be nice to print the AST and indicate where the failure
        # was?
        decorated_expr, line = add_error_marker(self.expression, self.expression_pos or -1)
        return """Error evaluating expression:\n{0}\n""".format(decorated_expr)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

import os
from pathlib import Path
from collections import namedtuple

import six

from scss.calculator import Calculator
from scss.compiler import Compiler
import scss.config as config
from scss.extension.bootstrap import BootstrapExtension
from scss.extension.core import CoreExtension
from scss.extension.compass import CompassExtension
from scss.extension.extra import ExtraExtension
from scss.extension.fonts import FontsExtension
from scss.namespace import Namespace
from scss.scss_meta import (
    BUILD_INFO, PROJECT, VERSION, REVISION, URL, AUTHOR, AUTHOR_EMAIL, LICENSE,
)
from scss.source import SourceFile
from scss.types import String


_default_scss_vars = {
    '$BUILD-INFO': String.unquoted(BUILD_INFO),
    '$PROJECT': String.unquoted(PROJECT),
    '$VERSION': String.unquoted(VERSION),
    '$REVISION': String.unquoted(REVISION),
    '$URL': String.unquoted(URL),
    '$AUTHOR': String.unquoted(AUTHOR),
    '$AUTHOR-EMAIL': String.unquoted(AUTHOR_EMAIL),
    '$LICENSE': String.unquoted(LICENSE),

    # unsafe chars will be hidden as vars
    '$--doubleslash': String.unquoted('//'),
    '$--bigcopen': String.unquoted('/*'),
    '$--bigcclose': String.unquoted('*/'),
    '$--doubledot': String.unquoted(':'),
    '$--semicolon': String.unquoted(';'),
    '$--curlybracketopen': String.unquoted('{'),
    '$--curlybracketclosed': String.unquoted('}'),
}

SourceFileTuple = namedtuple('SourceFileTuple', ('parent_dir', 'filename'))


# TODO using this should spew an actual deprecation warning
class Scss(object):
    """Original programmatic interface to the compiler.

    This class is now DEPRECATED.  See :mod:`scss.compiler` for the
    replacement.
    """
    def __init__(
            self, scss_vars=None, scss_opts=None, scss_files=None,
            super_selector='', live_errors=False,
            library=None, func_registry=None,
            search_paths=None):

        self.super_selector = super_selector

        self._scss_vars = {}
        if scss_vars:
            calculator = Calculator()
            for var_name, value in scss_vars.items():
                if isinstance(value, six.string_types):
                    scss_value = calculator.evaluate_expression(value)
                    if scss_value is None:
                        # TODO warning?
                        scss_value = String.unquoted(value)
                else:
                    scss_value = value
                self._scss_vars[var_name] = scss_value

        self._scss_opts = scss_opts or {}
        self._scss_files = scss_files
        self._library = func_registry or library
        self._search_paths = search_paths

        # If true, swallow compile errors and embed them in the output instead
        self.live_errors = live_errors

    def compile(
            self, scss_string=None, scss_file=None, source_files=None,
            super_selector=None, filename=None, is_sass=None,
            line_numbers=True, import_static_css=False):
        """Compile Sass to CSS.  Returns a single CSS string.

        This method is DEPRECATED; see :mod:`scss.compiler` instead.
        """
        # Derive our root namespace
        self.scss_vars = _default_scss_vars.copy()
        if self._scss_vars is not None:
            self.scss_vars.update(self._scss_vars)

        root_namespace = Namespace(
            variables=self.scss_vars,
            functions=self._library,
        )

        # Figure out search paths.  Fall back from provided explicitly to
        # defined globally to just searching the current directory
        search_paths = ['.']
        if self._search_paths is not None:
            assert not isinstance(self._search_paths, six.string_types), \
                "`search_paths` should be an iterable, not a string"
            search_paths.extend(self._search_paths)
        else:
            if config.LOAD_PATHS:
                if isinstance(config.LOAD_PATHS, six.string_types):
                    # Back-compat: allow comma-delimited
                    search_paths.extend(config.LOAD_PATHS.split(','))
                else:
                    search_paths.extend(config.LOAD_PATHS)

            search_paths.extend(self._scss_opts.get('load_paths', []))

        # Normalize a few old styles of options
        output_style = self._scss_opts.get('style', config.STYLE)
        if output_style is True:
            output_style = 'compressed'
        elif output_style is False:
            output_style = 'legacy'

        fixed_search_path = []
        for path in search_paths:
            if isinstance(path, six.string_types):
                fixed_search_path.append(Path(path))
            else:
                fixed_search_path.append(path)

        # Build the compiler
        compiler = Compiler(
            namespace=root_namespace,
            extensions=[
                CoreExtension,
                ExtraExtension,
                FontsExtension,
                CompassExtension,
                BootstrapExtension,
            ],
            search_path=fixed_search_path,
            import_static_css=import_static_css,
            live_errors=self.live_errors,
            generate_source_map=self._scss_opts.get('debug_info', False),
            output_style=output_style,
            warn_unused_imports=self._scss_opts.get('warn_unused', False),
            ignore_parse_errors=config.DEBUG,
            loops_have_own_scopes=config.CONTROL_SCOPING,
            undefined_variables_fatal=config.FATAL_UNDEFINED,
            super_selector=super_selector or self.super_selector,
        )
        # Gonna add the source files manually
        compilation = compiler.make_compilation()

        # Inject the files we know about
        # TODO how does this work with the expectation of absoluteness
        if source_files is not None:
            for source in source_files:
                compilation.add_source(source)
        elif scss_string is not None:
            source = SourceFile.from_string(
                scss_string,
                relpath=filename,
                is_sass=is_sass,
            )
            compilation.add_source(source)
        elif scss_file is not None:
            # This is now the only way to allow forcibly overriding the
            # filename a source "thinks" it is
            with open(scss_file, 'rb') as f:
                source = SourceFile.from_file(
                    f,
                    relpath=filename or scss_file,
                    is_sass=is_sass,
                )
            compilation.add_source(source)

        # Plus the ones from the constructor
        if self._scss_files:
            for name, contents in list(self._scss_files.items()):
                source = SourceFile.from_string(contents, relpath=name)
                compilation.add_source(source)

        compiled = compiler.call_and_catch_errors(compilation.run)
        self.source_files = list(SourceFileTuple(*os.path.split(s.path)) for s in compilation.source_index.values())
        return compiled

    # Old, old alias
    Compilation = compile

    def get_scss_constants(self):
        scss_vars = self.root_namespace.variables
        return dict(
            (k, v) for k, v in scss_vars.items()
            if k and (not k.startswith('$') or k[1].isupper())
        )

    def get_scss_vars(self):
        scss_vars = self.root_namespace.variables
        return dict(
            (k, v) for k, v in scss_vars.items()
            if k and not (not k.startswith('$') and k[1].isupper())
        )



"""Functions used for generating packed CSS sprite maps.

These are ported from the Binary Tree Bin Packing Algorithm:
http://codeincomplete.com/posts/2011/5/7/bin_packing/
"""
from __future__ import absolute_import
from __future__ import unicode_literals

# Copyright (c) 2011, 2012, 2013 Jake Gordon and contributors
# Copyright (c) 2013 German M. Bravo

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class LayoutNode(object):
    def __init__(self, x, y, w, h, down=None, right=None, used=False):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.down = down
        self.right = right
        self.used = used
        self.width = 0
        self.height = 0

    @property
    def area(self):
        return self.width * self.height

    def __repr__(self):
        return '<%s (%s, %s) [%sx%s]>' % (self.__class__.__name__, self.x, self.y, self.w, self.h)


class SpritesLayout(object):
    def __init__(self, blocks, padding=None, margin=None, ppadding=None, pmargin=None):
        self.num_blocks = len(blocks)

        if margin is None:
            margin = [[0] * 4] * self.num_blocks
        elif not isinstance(margin, (tuple, list)):
            margin = [[margin] * 4] * self.num_blocks
        elif not isinstance(margin[0], (tuple, list)):
            margin = [margin] * self.num_blocks

        if padding is None:
            padding = [[0] * 4] * self.num_blocks
        elif not isinstance(padding, (tuple, list)):
            padding = [[padding] * 4] * self.num_blocks
        elif not isinstance(padding[0], (tuple, list)):
            padding = [padding] * self.num_blocks

        if pmargin is None:
            pmargin = [[0.0] * 4] * self.num_blocks
        elif not isinstance(pmargin, (tuple, list)):
            pmargin = [[pmargin] * 4] * self.num_blocks
        elif not isinstance(pmargin[0], (tuple, list)):
            pmargin = [pmargin] * self.num_blocks

        if ppadding is None:
            ppadding = [[0.0] * 4] * self.num_blocks
        elif not isinstance(ppadding, (tuple, list)):
            ppadding = [[ppadding] * 4] * self.num_blocks
        elif not isinstance(ppadding[0], (tuple, list)):
            ppadding = [ppadding] * self.num_blocks

        self.blocks = tuple((
            b[0] + padding[i][3] + padding[i][1] + margin[i][3] + margin[i][1] + int(round(b[0] * (ppadding[i][3] + ppadding[i][1] + pmargin[i][3] + pmargin[i][1]))),
            b[1] + padding[i][0] + padding[i][2] + margin[i][0] + margin[i][2] + int(round(b[1] * (ppadding[i][0] + ppadding[i][2] + pmargin[i][0] + pmargin[i][2]))),
            b[0],
            b[1],
            i
        ) for i, b in enumerate(blocks))

        self.margin = margin
        self.padding = padding
        self.pmargin = pmargin
        self.ppadding = ppadding


class PackedSpritesLayout(SpritesLayout):
    @staticmethod
    def MAXSIDE(a, b):
        """maxside: Sort pack by maximum sides"""
        return cmp(max(b[0], b[1]), max(a[0], a[1])) or cmp(min(b[0], b[1]), min(a[0], a[1])) or cmp(b[1], a[1]) or cmp(b[0], a[0])

    @staticmethod
    def WIDTH(a, b):
        """width: Sort pack by width"""
        return cmp(b[0], a[0]) or cmp(b[1], a[1])

    @staticmethod
    def HEIGHT(a, b):
        """height: Sort pack by height"""
        return cmp(b[1], a[1]) or cmp(b[0], a[0])

    @staticmethod
    def AREA(a, b):
        """area: Sort pack by area"""
        return cmp(b[0] * b[1], a[0] * a[1]) or cmp(b[1], a[1]) or cmp(b[0], a[0])

    def __init__(self, blocks, padding=None, margin=None, ppadding=None, pmargin=None, methods=None):
        super(PackedSpritesLayout, self).__init__(blocks, padding, margin, ppadding, pmargin)

        ratio = 0

        if methods is None:
            methods = (self.MAXSIDE, self.WIDTH, self.HEIGHT, self.AREA)

        for method in methods:
            sorted_blocks = sorted(
                self.blocks,
                cmp=method,
            )
            root = LayoutNode(
                x=0,
                y=0,
                w=sorted_blocks[0][0] if sorted_blocks else 0,
                h=sorted_blocks[0][1] if sorted_blocks else 0
            )

            area = 0
            nodes = [None] * self.num_blocks

            for block in sorted_blocks:
                w, h, width, height, i = block
                node = self._findNode(root, w, h)
                if node:
                    node = self._splitNode(node, w, h)
                else:
                    root = self._growNode(root, w, h)
                    node = self._findNode(root, w, h)
                    if node:
                        node = self._splitNode(node, w, h)
                    else:
                        node = None
                nodes[i] = node
                node.width = width
                node.height = height
                area += node.area

            this_ratio = area / float(root.w * root.h)
            # print method.__doc__, "%g%%" % (this_ratio * 100)
            if ratio < this_ratio:
                self.root = root
                self.nodes = nodes
                self.method = method
                ratio = this_ratio
                if ratio > 0.96:
                    break
        # print self.method.__doc__, "%g%%" % (ratio * 100)

    def __iter__(self):
        for i, node in enumerate(self.nodes):
            margin, padding = self.margin[i], self.padding[i]
            pmargin, ppadding = self.pmargin[i], self.ppadding[i]
            cssw = node.width + padding[3] + padding[1] + int(round(node.width * (ppadding[3] + ppadding[1])))  # image width plus padding
            cssh = node.height + padding[0] + padding[2] + int(round(node.height * (ppadding[0] + ppadding[2])))  # image height plus padding
            cssx = node.x + margin[3] + int(round(node.width * pmargin[3]))
            cssy = node.y + margin[0] + int(round(node.height * pmargin[0]))
            x = cssx + padding[3] + int(round(node.width * ppadding[3]))
            y = cssy + padding[0] + int(round(node.height * ppadding[0]))
            yield x, y, node.width, node.height, cssx, cssy, cssw, cssh

    @property
    def width(self):
        return self.root.w

    @property
    def height(self):
        return self.root.h

    def _findNode(self, root, w, h):
        if root.used:
            return self._findNode(root.right, w, h) or self._findNode(root.down, w, h)
        elif w <= root.w and h <= root.h:
            return root
        else:
            return None

    def _splitNode(self, node, w, h):
        node.used = True
        node.down = LayoutNode(
            x=node.x,
            y=node.y + h,
            w=node.w,
            h=node.h - h
        )
        node.right = LayoutNode(
            x=node.x + w,
            y=node.y,
            w=node.w - w,
            h=h
        )
        return node

    def _growNode(self, root, w, h):
        canGrowDown = w <= root.w
        canGrowRight = h <= root.h

        shouldGrowRight = canGrowRight and (root.h >= root.w + w)  # attempt to keep square-ish by growing right when height is much greater than width
        shouldGrowDown = canGrowDown and (root.w >= root.h + h)  # attempt to keep square-ish by growing down when width is much greater than height

        if shouldGrowRight:
            return self._growRight(root, w, h)
        elif shouldGrowDown:
            return self._growDown(root, w, h)
        elif canGrowRight:
            return self._growRight(root, w, h)
        elif canGrowDown:
            return self._growDown(root, w, h)
        else:
            # need to ensure sensible root starting size to avoid this happening
            assert False, "Blocks must be properly sorted!"

    def _growRight(self, root, w, h):
        root = LayoutNode(
            used=True,
            x=0,
            y=0,
            w=root.w + w,
            h=root.h,
            down=root,
            right=LayoutNode(
                x=root.w,
                y=0,
                w=w,
                h=root.h
            )
        )
        return root

    def _growDown(self, root, w, h):
        root = LayoutNode(
            used=True,
            x=0,
            y=0,
            w=root.w,
            h=root.h + h,
            down=LayoutNode(
                x=0,
                y=root.h,
                w=root.w,
                h=h
            ),
            right=root
        )
        return root


class HorizontalSpritesLayout(SpritesLayout):
    def __init__(self, blocks, padding=None, margin=None, ppadding=None, pmargin=None, position=None):
        super(HorizontalSpritesLayout, self).__init__(blocks, padding, margin, ppadding, pmargin)

        self.width = sum(block[0] for block in self.blocks)
        self.height = max(block[1] for block in self.blocks)

        if position is None:
            position = [0.0] * self.num_blocks
        elif not isinstance(position, (tuple, list)):
            position = [position] * self.num_blocks
        self.position = position

    def __iter__(self):
        cx = 0
        for i, block in enumerate(self.blocks):
            w, h, width, height, i = block
            margin, padding = self.margin[i], self.padding[i]
            pmargin, ppadding = self.pmargin[i], self.ppadding[i]
            position = self.position[i]
            cssw = width + padding[3] + padding[1] + int(round(width * (ppadding[3] + ppadding[1])))  # image width plus padding
            cssh = height + padding[0] + padding[2] + int(round(height * (ppadding[0] + ppadding[2])))  # image height plus padding
            cssx = cx + margin[3] + int(round(width * pmargin[3]))  # anchored at x
            cssy = int(round((self.height - cssh) * position))  # centered vertically
            x = cssx + padding[3] + int(round(width * ppadding[3]))  # image drawn offset to account for padding
            y = cssy + padding[0] + int(round(height * ppadding[0]))  # image drawn offset to account for padding
            yield x, y, width, height, cssx, cssy, cssw, cssh
            cx += cssw + margin[3] + margin[1] + int(round(width * (pmargin[3] + pmargin[1])))


class VerticalSpritesLayout(SpritesLayout):
    def __init__(self, blocks, padding=None, margin=None, ppadding=None, pmargin=None, position=None):
        super(VerticalSpritesLayout, self).__init__(blocks, padding, margin, ppadding, pmargin)

        self.width = max(block[0] for block in self.blocks)
        self.height = sum(block[1] for block in self.blocks)

        if position is None:
            position = [0.0] * self.num_blocks
        elif not isinstance(position, (tuple, list)):
            position = [position] * self.num_blocks
        self.position = position

    def __iter__(self):
        cy = 0
        for i, block in enumerate(self.blocks):
            w, h, width, height, i = block
            margin, padding = self.margin[i], self.padding[i]
            pmargin, ppadding = self.pmargin[i], self.ppadding[i]
            position = self.position[i]
            cssw = width + padding[3] + padding[1] + int(round(width * (ppadding[3] + ppadding[1])))  # image width plus padding
            cssh = height + padding[0] + padding[2] + int(round(height * (ppadding[0] + ppadding[2])))  # image height plus padding
            cssx = int(round((self.width - cssw) * position))  # centered horizontally
            cssy = cy + margin[0] + int(round(height * pmargin[0]))  # anchored at y
            x = cssx + padding[3] + int(round(width * ppadding[3]))  # image drawn offset to account for padding
            y = cssy + padding[0] + int(round(height * ppadding[0]))  # image drawn offset to account for padding
            yield x, y, width, height, cssx, cssy, cssw, cssh
            cy += cssh + margin[0] + margin[2] + int(round(height * (pmargin[0] + pmargin[2])))


class DiagonalSpritesLayout(SpritesLayout):
    def __init__(self, blocks, padding=None, margin=None, ppadding=None, pmargin=None):
        super(DiagonalSpritesLayout, self).__init__(blocks, padding, margin, ppadding, pmargin)
        self.width = sum(block[0] for block in self.blocks)
        self.height = sum(block[1] for block in self.blocks)

    def __iter__(self):
        cx, cy = 0, 0
        for i, block in enumerate(self.blocks):
            w, h, width, height, i = block
            margin, padding = self.margin[i], self.padding[i]
            pmargin, ppadding = self.pmargin[i], self.ppadding[i]
            cssw = width + padding[3] + padding[1] + int(round(width * (ppadding[3] + ppadding[1])))  # image width plus padding
            cssh = height + padding[0] + padding[2] + int(round(height * (ppadding[0] + ppadding[2])))  # image height plus padding
            cssx = cx + margin[3] + int(round(width * pmargin[3]))  # anchored at x
            cssy = cy + margin[0] + int(round(height * pmargin[0]))  # anchored at y
            x = cssx + padding[3] + int(round(width * ppadding[3]))  # image drawn offset to account for padding
            y = cssy + padding[0] + int(round(height * ppadding[0]))  # image drawn offset to account for padding
            yield x, y, width, height, cssx, cssy, cssw, cssh
            cx += cssw + margin[3] + margin[1] + int(round(width * (pmargin[3] + pmargin[1])))
            cy += cssh + margin[0] + margin[2] + int(round(height * (pmargin[0] + pmargin[2])))
